# degu: an embedded HTTP server and client library
# Copyright (C) 2014 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Unit tests for the `degu.base` module`
"""

from unittest import TestCase
import os
import io
import sys
from random import SystemRandom
import itertools

from . import helpers
from .helpers import DummySocket, random_chunks, FuzzTestCase
from degu.sslhelpers import random_id
from degu.base import _MAX_LINE_SIZE
from degu import base, _basepy


# True if the C extension is available
try:
    from degu import _base
    C_EXT_AVAIL = True
except ImportError:
    _base = None
    C_EXT_AVAIL = False


random = SystemRandom()


BAD_HEADER_LINES = (
    b'K:V\r\n',
    b'K V\r\n',
    b': V\r\n',
    b'K: \r\n',
    b': \r\n',
)

GOOD_HEADERS = (
    (
        b'Content-Type: application/json\r\n',
        ('content-type', 'application/json')
    ),
    (
        b'Content-Length: 17\r\n',
        ('content-length', 17)
    ),
    (
        b'Content-Length: 0\r\n',
        ('content-length', 0)
    ),
    (
        b'Transfer-Encoding: chunked\r\n',
        ('transfer-encoding', 'chunked')
    ),
)


def _permute_remove(method):
    if len(method) <= 1:
        return
    for i in range(len(method)):
        m = bytearray(method)
        del m[i]
        m = bytes(m)
        yield m
        yield from _permute_remove(m)


def _permute_replace(method):
    for i in range(len(method)):
        for j in range(256):
            if method[i] == j:
                continue
            m = bytearray(method)
            m[i] = j
            yield bytes(m)


def _permute_insert(method):
    for i in range(len(method) + 1):
        for j in range(256):
            m = bytearray(method)
            m.insert(i, j)
            yield bytes(m)


GOOD_METHODS = (
    'GET',
    'HEAD',
    'POST',
    'PUT',
    'DELETE',
)
_functions = (_permute_remove, _permute_replace, _permute_insert)
BAD_METHODS = [
    b'',
    b'TRACE',
    b'OPTIONS',
    b'CONNECT',
    b'PATCH',
]
BAD_METHODS.extend(m.encode().lower() for m in GOOD_METHODS)
for func in _functions:
    for m in GOOD_METHODS:
        BAD_METHODS.extend(func(m.encode()))
BAD_METHODS = tuple(sorted(set(BAD_METHODS)))


def random_headers(count):
    return dict(
        ('X-' + random_id(), random_id()) for i in range(count)
    )


def build_header_lines(headers):
    return ''.join(
        '{}: {}\r\n'.format(key, value) for (key, value) in headers.items()
    ).encode('latin_1')


def casefold_headers(headers):
    """
    For example:

    >>> casefold_headers({'FOO': 'BAR'})
    {'foo': 'BAR'}

    """
    return dict(
        (key.casefold(), value) for (key, value) in headers.items()
    )


def random_line():
    return '{}\r\n'.format(random_id()).encode()


def random_header_line():
    return '{}: {}\r\n'.format(random_id(), random_id()).encode()


def random_lines(header_count=15):
    first_line = random_id()
    header_lines = [random_id() for i in range(header_count)]
    return (first_line, header_lines)


def encode_preamble(first_line, header_lines):
    lines = [first_line + '\r\n']
    lines.extend(line + '\r\n' for line in header_lines)
    lines.append('\r\n')
    return ''.join(lines).encode('latin_1')


def random_body():
    size = random.randint(1, 34969)
    return os.urandom(size)


class AlternatesTestCase(FuzzTestCase):
    def skip_if_no_c_ext(self):
        if not C_EXT_AVAIL:
            self.skipTest('cannot import `degu._base` C extension')


MiB = 1024 * 1024

class TestConstants(TestCase):
    def check_power_of_two(self, name, size):
        self.assertIsInstance(size, int, name)
        self.assertGreaterEqual(size, 1024, name)
        self.assertLessEqual(size, MiB * 32, name)
        self.assertFalse(size & (size - 1),
            '({}) {:d} is not a power of 2'.format(name, size)
        )

    def check_size_constant(self, name, min_size=4096, max_size=16777216):
        self.check_power_of_two('min_size', min_size)
        self.check_power_of_two('max_size', max_size)
        self.assertEqual(name[-5:], '_SIZE', name)
        self.assertTrue(name.isupper(), '{!r} not uppercase'.format(name))
        size = getattr(base, name)
        self.check_power_of_two(name, size)
        self.assertGreaterEqual(size, min_size, name)
        self.assertLessEqual(size, max_size, name)

    def test__MAX_LINE_SIZE(self):
        self.assertIsInstance(base._MAX_LINE_SIZE, int)
        self.assertGreaterEqual(base._MAX_LINE_SIZE, 1024)
        self.assertEqual(base._MAX_LINE_SIZE % 1024, 0)
        self.assertLessEqual(base._MAX_LINE_SIZE, 8192)

    def test__MAX_HEADER_COUNT(self):
        self.assertIsInstance(base._MAX_HEADER_COUNT, int)
        self.assertGreaterEqual(base._MAX_HEADER_COUNT, 5)
        self.assertLessEqual(base._MAX_HEADER_COUNT, 20)

    def test_STREAM_BUFFER_SIZE(self):
        self.assertIsInstance(base.STREAM_BUFFER_SIZE, int)
        self.assertEqual(base.STREAM_BUFFER_SIZE % 4096, 0)
        self.assertGreaterEqual(base.STREAM_BUFFER_SIZE, 4096)

    def test_MAX_READ_SIZE(self):
        self.check_size_constant('MAX_READ_SIZE')

    def test_MAX_CHUNK_SIZE(self):
        self.check_size_constant('MAX_CHUNK_SIZE')

    def test_IO_SIZE(self):
        self.check_size_constant('IO_SIZE')

    def test_bodies(self):
        self.assertTrue(issubclass(base.BodiesAPI, tuple))
        self.assertIsInstance(base.bodies, tuple)
        self.assertIsInstance(base.bodies, base.BodiesAPI)

        self.assertIs(base.bodies.Body, base.Body)
        self.assertIs(base.bodies.BodyIter, base.BodyIter)
        self.assertIs(base.bodies.ChunkedBody, base.ChunkedBody)
        self.assertIs(base.bodies.ChunkedBodyIter, base.ChunkedBodyIter)

        self.assertIs(base.bodies[0], base.Body)
        self.assertIs(base.bodies[1], base.BodyIter)
        self.assertIs(base.bodies[2], base.ChunkedBody)
        self.assertIs(base.bodies[3], base.ChunkedBodyIter)

        self.assertEqual(base.bodies,
            (base.Body, base.BodyIter, base.ChunkedBody, base.ChunkedBodyIter)
        )


class TestEmptyPreambleError(TestCase):
    def test_init(self):
        e = base.EmptyPreambleError('stuff and junk')
        self.assertIsInstance(e, Exception)
        self.assertIsInstance(e, ConnectionError)
        self.assertIs(type(e), base.EmptyPreambleError)
        self.assertEqual(str(e), 'stuff and junk')


class FuzzTestFunctions(AlternatesTestCase):
    def test__read_preamble_p(self):
        self.fuzz(_basepy._read_preamble)

    def test__read_preamble_c(self):
        self.skip_if_no_c_ext()
        self.fuzz(_base._read_preamble)

    def test_read_chunk(self):
        self.fuzz(base.read_chunk)


class DummyFile:
    def __init__(self, lines):
        self._lines = lines
        self._calls = []

    def readline(self, size=None):
        self._calls.append(size)
        return self._lines.pop(0)


class DummyWriter:
    def __init__(self):
        self._calls = []

    def write(self, data):
        assert isinstance(data, bytes)
        self._calls.append(('write', data))
        return len(data)

    def flush(self):
        self._calls.append('flush')


class UserBytes(bytes):
    pass


class TestFunctions(AlternatesTestCase):
    def test__makefiles(self):
        sock = DummySocket()
        self.assertEqual(base._makefiles(sock), (sock._rfile, sock._wfile))
        self.assertEqual(sock._calls, [
            ('makefile', 'rb', {'buffering': base.STREAM_BUFFER_SIZE}),
            ('makefile', 'wb', {'buffering': base.STREAM_BUFFER_SIZE}),
        ])

    def check_parse_method(self, backend):
        self.assertIn(backend, (_base, _basepy))
        parse_method = backend.parse_method

        for method in GOOD_METHODS:
            expected = getattr(backend, method)

            # Input is str:
            result = parse_method(method)
            self.assertEqual(result, method)
            self.assertIs(result, expected)

            # Input is bytes:
            result = parse_method(method.encode())
            self.assertEqual(result, method)
            self.assertIs(result, expected)

            # Lowercase str:
            with self.assertRaises(ValueError) as cm:
                parse_method(method.lower())
            self.assertEqual(str(cm.exception),
                'bad HTTP method: {!r}'.format(method.lower().encode())
            )

            # Lowercase bytes:
            with self.assertRaises(ValueError) as cm:
                parse_method(method.lower().encode())
            self.assertEqual(str(cm.exception),
                'bad HTTP method: {!r}'.format(method.lower().encode())
            )

        # Static bad methods:
        bad_methods = (
            'OPTIONS',
            'TRACE',
            'CONNECT',
            'FOO',
            'BAR',
            'COPY',
            'FOUR',
            'SIXSIX',
            'FOOBAR',
            '',
        )
        for bad in bad_methods:
            # Bad str:
            with self.assertRaises(ValueError) as cm:
                parse_method(bad)
            self.assertEqual(str(cm.exception),
                'bad HTTP method: {!r}'.format(bad.encode())
            )

            # Bad bytes:
            with self.assertRaises(ValueError) as cm:
                parse_method(bad.encode())
            self.assertEqual(str(cm.exception),
                'bad HTTP method: {!r}'.format(bad.encode())
            )

        # Pre-generated bad method permutations:
        for bad in BAD_METHODS:
            with self.assertRaises(ValueError) as cm:
                parse_method(bad)
            self.assertEqual(str(cm.exception),
                'bad HTTP method: {!r}'.format(bad)
            )

        # Random bad bytes:
        for size in range(1, 20):
            for i in range(100):
                bad = os.urandom(size)
                with self.assertRaises(ValueError) as cm:
                    parse_method(bad)
                self.assertEqual(str(cm.exception),
                    'bad HTTP method: {!r}'.format(bad)
                )

    def test_parse_method_py(self):
        self.check_parse_method(_basepy)

    def test_parse_method_c(self):
        self.skip_if_no_c_ext()
        self.check_parse_method(_base)

    def check_parse_response_line(self, backend):
        self.assertIn(backend, (_base, _basepy))
        parse_response_line = backend.parse_response_line

        # request line is too short:
        line  = b'HTTP/1.1 200 OK'
        for stop in range(15):
            short = line[:stop]
            self.assertTrue(0 <= len(short) <= 14)
            with self.assertRaises(ValueError) as cm:
                parse_response_line(short)
            self.assertEqual(str(cm.exception),
                'response line too short: {!r}'.format(short)
            )

        # Double confirm when len(line) is 0:
        with self.assertRaises(ValueError) as cm:
            parse_response_line(b'')
        self.assertEqual(str(cm.exception),
            "response line too short: b''"
        )

        # Double confirm when len(line) is 14:
        short = line[:14]
        self.assertEqual(len(short), 14)
        with self.assertRaises(ValueError) as cm:
            parse_response_line(short)
        self.assertEqual(str(cm.exception),
            "response line too short: b'HTTP/1.1 200 O'"
        )

        # Confirm valid minimum response line is 15 bytes in length:
        self.assertEqual(len(line), 15)
        self.assertEqual(parse_response_line(line), (200, 'OK'))

        # Test all status in range 000-999, plus a few valid reasons:
        for status in range(1000):
            for reason in ('OK', 'Not Found', 'Enhance Your Calm'):
                line = 'HTTP/1.1 {:03d} {}'.format(status, reason).encode()
                if 100 <= status <= 599:
                    self.assertEqual(parse_response_line(line), (status, reason))
                else:
                    with self.assertRaises(ValueError) as cm:
                        parse_response_line(line)
                    self.assertEqual(str(cm.exception),
                        'bad status in response line: {!r}'.format(line)
                    )

        # Test fast-path when reason is 'OK':
        for status in range(200, 600):
            line = 'HTTP/1.1 {} OK'.format(status).encode()
            tup = parse_response_line(line)
            self.assertEqual(tup, (status, 'OK'))
            self.assertIs(tup[1], backend.OK)

        # Permutations:
        good = b'HTTP/1.1 200 OK'
        self.assertEqual(parse_response_line(good), (200, 'OK'))
        for i in range(len(good)):
            bad = bytearray(good)
            del bad[i]
            with self.assertRaises(ValueError):
                parse_response_line(bytes(bad))
            for j in range(32):
                bad = bytearray(good)
                bad[i] = j
                with self.assertRaises(ValueError):
                    parse_response_line(bytes(bad))

    def test_parse_response_line_py(self):
        self.check_parse_response_line(_basepy)

    def test_parse_response_line_c(self):
        self.skip_if_no_c_ext()
        self.check_parse_response_line(_base)

    def check_parse_request_line(self, backend):
        self.assertIn(backend, (_base, _basepy))
        parse_request_line = backend.parse_request_line
        good_uri = ('/foo', '/?stuff=junk', '/foo/bar/', '/foo/bar?stuff=junk')

        # Test all good methods:
        for method in GOOD_METHODS:
            good = '{} / HTTP/1.1'.format(method).encode()
            self.assertEqual(parse_request_line(good), (method, '/'))
            for i in range(len(good)):
                bad = bytearray(good)
                del bad[i]
                with self.assertRaises(ValueError):
                    parse_request_line(bytes(bad))
                for j in range(256):
                    if good[i] == j:
                        continue
                    bad = bytearray(good)
                    bad[i] = j
                    with self.assertRaises(ValueError):
                        parse_request_line(bytes(bad))
            for uri in good_uri:
                good2 = '{} {} HTTP/1.1'.format(method, uri).encode()
                self.assertEqual(parse_request_line(good2), (method, uri))

        # Pre-generated bad method permutations:
        for uri in good_uri:
            tail = ' {} HTTP/1.1'.format(uri).encode()
            for method in BAD_METHODS:
                bad = method + tail
                with self.assertRaises(ValueError):
                    parse_request_line(bad)

    def test_parse_request_line_py(self):
        self.check_parse_request_line(_basepy)

    def test_parse_request_line_c(self):
        self.skip_if_no_c_ext()
        self.check_parse_request_line(_base)

    def check_format_request_preamble(self, backend):
        # Too few arguments:
        with self.assertRaises(TypeError):
            backend.format_request_preamble()
        with self.assertRaises(TypeError):
            backend.format_request_preamble('GET')
        with self.assertRaises(TypeError):
            backend.format_request_preamble('GET', '/foo')

        # Too many arguments:
        with self.assertRaises(TypeError):
            backend.format_request_preamble('GET', '/foo', {}, None)

        # No headers:
        self.assertEqual(
            backend.format_request_preamble('GET', '/foo', {}),
            b'GET /foo HTTP/1.1\r\n\r\n'
        )

        # One header:
        headers = {'content-length': 1776}
        self.assertEqual(
            backend.format_request_preamble('PUT', '/foo', headers),
            b'PUT /foo HTTP/1.1\r\ncontent-length: 1776\r\n\r\n'
        )
        headers = {'transfer-encoding': 'chunked'}
        self.assertEqual(
            backend.format_request_preamble('POST', '/foo', headers),
            b'POST /foo HTTP/1.1\r\ntransfer-encoding: chunked\r\n\r\n'
        )

        # Two headers:
        headers = {'content-length': 1776, 'a': 'A'}
        self.assertEqual(
            backend.format_request_preamble('PUT', '/foo', headers),
            b'PUT /foo HTTP/1.1\r\na: A\r\ncontent-length: 1776\r\n\r\n'
        )
        headers = {'transfer-encoding': 'chunked', 'z': 'Z'}
        self.assertEqual(
            backend.format_request_preamble('POST', '/foo', headers),
            b'POST /foo HTTP/1.1\r\ntransfer-encoding: chunked\r\nz: Z\r\n\r\n'
        )

        # Three headers:
        headers = {'content-length': 1776, 'a': 'A', 'z': 'Z'}
        self.assertEqual(
            backend.format_request_preamble('PUT', '/foo', headers),
            b'PUT /foo HTTP/1.1\r\na: A\r\ncontent-length: 1776\r\nz: Z\r\n\r\n'
        )
        headers = {'transfer-encoding': 'chunked', 'z': 'Z', 'a': 'A'}
        self.assertEqual(
            backend.format_request_preamble('POST', '/foo', headers),
            b'POST /foo HTTP/1.1\r\na: A\r\ntransfer-encoding: chunked\r\nz: Z\r\n\r\n'
        )

    def test_format_request_preamble_py(self):
        self.check_format_request_preamble(_basepy)

    def test_format_request_preamble_c(self):
        self.skip_if_no_c_ext()
        self.check_format_request_preamble(_base)

    def check_format_response_preamble(self, backend):
        # Too few arguments:
        with self.assertRaises(TypeError):
            backend.format_response_preamble()
        with self.assertRaises(TypeError):
            backend.format_response_preamble(200)
        with self.assertRaises(TypeError):
            backend.format_response_preamble(200, 'OK')

        # Too many arguments:
        with self.assertRaises(TypeError):
            backend.format_response_preamble('200', 'OK', {}, None)

        # No headers:
        self.assertEqual(
            backend.format_response_preamble(200, 'OK', {}),
            b'HTTP/1.1 200 OK\r\n\r\n'
        )

        # One header:
        headers = {'content-length': 1776}
        self.assertEqual(
            backend.format_response_preamble(200, 'OK', headers),
            b'HTTP/1.1 200 OK\r\ncontent-length: 1776\r\n\r\n'
        )
        headers = {'transfer-encoding': 'chunked'}
        self.assertEqual(
            backend.format_response_preamble(200, 'OK', headers),
            b'HTTP/1.1 200 OK\r\ntransfer-encoding: chunked\r\n\r\n'
        )

        # Two headers:
        headers = {'content-length': 1776, 'a': 'A'}
        self.assertEqual(
            backend.format_response_preamble(200, 'OK', headers),
            b'HTTP/1.1 200 OK\r\na: A\r\ncontent-length: 1776\r\n\r\n'
        )
        headers = {'transfer-encoding': 'chunked', 'z': 'Z'}
        self.assertEqual(
            backend.format_response_preamble(200, 'OK', headers),
            b'HTTP/1.1 200 OK\r\ntransfer-encoding: chunked\r\nz: Z\r\n\r\n'
        )

        # Three headers:
        headers = {'content-length': 1776, 'a': 'A', 'z': 'Z'}
        self.assertEqual(
            backend.format_response_preamble(200, 'OK', headers),
            b'HTTP/1.1 200 OK\r\na: A\r\ncontent-length: 1776\r\nz: Z\r\n\r\n'
        )
        headers = {'transfer-encoding': 'chunked', 'z': 'Z', 'a': 'A'}
        self.assertEqual(
            backend.format_response_preamble(200, 'OK', headers),
            b'HTTP/1.1 200 OK\r\na: A\r\ntransfer-encoding: chunked\r\nz: Z\r\n\r\n'
        )

    def test_format_response_preamble_py(self):
        self.check_format_response_preamble(_basepy)

    def test_format_response_preamble_c(self):
        self.skip_if_no_c_ext()
        self.check_format_response_preamble(_base)

    def check_parse_preamble(self, backend):
        self.assertEqual(backend.parse_preamble(b'Foo'), ('Foo', {}))
        parse_preamble = backend.parse_preamble

        self.assertEqual(backend.parse_preamble(b'Foo\r\nBar: Baz'),
            ('Foo', {'bar': 'Baz'})
        )
        self.assertEqual(backend.parse_preamble(b'Foo\r\nContent-Length: 42'),
            ('Foo', {'content-length': 42})
        )
        self.assertEqual(
            backend.parse_preamble(b'Foo\r\nTransfer-Encoding: chunked'),
            ('Foo', {'transfer-encoding': 'chunked'})
        )

        # Bad bytes in first line:
        with self.assertRaises(ValueError) as cm:
            backend.parse_preamble(b'Foo\x00\r\nBar: Baz')
        self.assertEqual(str(cm.exception),
            "bad bytes in first line: b'Foo\\x00'"
        )

        # Bad bytes in header name:
        with self.assertRaises(ValueError) as cm:
            backend.parse_preamble(b'Foo\r\nBar\x00: Baz')
        self.assertEqual(str(cm.exception),
            "bad bytes in header name: b'Bar\\x00'"
        )

        # Bad bytes in header value:
        with self.assertRaises(ValueError) as cm:
            backend.parse_preamble(b'Foo\r\nBar: Baz\x00')
        self.assertEqual(str(cm.exception),
            "bad bytes in header value: b'Baz\\x00'"
        )

        # content-length larger than 9007199254740992:
        value = 9007199254740992
        for gv in (0, 17, value, value - 1, value - 17):
            buf = 'GET / HTTP/1.1\r\nContent-Length: {:d}'.format(gv).encode()
            self.assertEqual(parse_preamble(buf),
                ('GET / HTTP/1.1', {'content-length': gv})
            )
        with self.assertRaises(ValueError) as cm:
            parse_preamble(b'GET / HTTP/1.1\r\nContent-Length: 09007199254740992')
        self.assertEqual(str(cm.exception),
            "content-length too long: b'0900719925474099'..."
        )
        for i in range(1, 101):
            bv = value + i
            buf = 'GET / HTTP/1.1\r\nContent-Length: {:d}'.format(bv).encode()
            with self.assertRaises(ValueError) as cm:
                backend.parse_preamble(buf)
            self.assertEqual(str(cm.exception),
                'content-length value too large: {:d}'.format(bv)
            )
        buf = b'GET / HTTP/1.1\r\nContent-Length: 9999999999999999'
        with self.assertRaises(ValueError) as cm:
            backend.parse_preamble(buf)
        self.assertEqual(str(cm.exception),
            'content-length value too large: 9999999999999999'
        )

    def test_parse_preamble_p(self):
        self.check_parse_preamble(_basepy)

    def test_parse_preamble_c(self):
        self.skip_if_no_c_ext()
        self.check_parse_preamble(_base)

    def check__read_preamble(self, backend):
        self.assertIn(backend, (_basepy, _base))
        read_preamble = backend._read_preamble

        # Bad bytes in preamble first line:
        for size in range(1, 8):
            for bad in helpers.iter_bad_values(size):
                data = bad + b'\r\nFoo: Bar\r\nstuff: Junk\r\n\r\n'
                rfile = io.BytesIO(data)
                with self.assertRaises(ValueError) as cm:
                    read_preamble(rfile)
                self.assertEqual(str(cm.exception),
                    'bad bytes in first line: {!r}'.format(bad)
                )
                self.assertEqual(sys.getrefcount(rfile), 2)
                self.assertEqual(rfile.tell(), size + 2)

        # Bad bytes in header name:
        for size in range(1, 8):
            for bad in helpers.iter_bad_keys(size):
                data = b'da first line\r\n' + bad + b': Bar\r\nstuff: Junk\r\n\r\n'
                rfile = io.BytesIO(data)
                with self.assertRaises(ValueError) as cm:
                    read_preamble(rfile)
                self.assertEqual(str(cm.exception),
                    'bad bytes in header name: {!r}'.format(bad)
                )
                self.assertEqual(sys.getrefcount(rfile), 2)
                self.assertEqual(rfile.tell(), size + 22)

        # Bad bytes in header value:
        for size in range(1, 8):
            for bad in helpers.iter_bad_values(size):
                data = b'da first line\r\nFoo: ' + bad + b'\r\nstuff: Junk\r\n\r\n'
                rfile = io.BytesIO(data)
                with self.assertRaises(ValueError) as cm:
                    read_preamble(rfile)
                self.assertEqual(str(cm.exception),
                    'bad bytes in header value: {!r}'.format(bad)
                )
                self.assertEqual(sys.getrefcount(rfile), 2)
                self.assertEqual(rfile.tell(), size + 22)

        # Test number of arguments _read_preamble() takes:
        with self.assertRaises(TypeError) as cm:
            read_preamble()
        self.assertIn(str(cm.exception), {
            '_read_preamble() takes exactly 1 argument (0 given)',
            "_read_preamble() missing 1 required positional argument: 'rfile'"
        })
        with self.assertRaises(TypeError) as cm:
            read_preamble('foo', 'bar')
        self.assertIn(str(cm.exception), {
            '_read_preamble() takes exactly 1 argument (2 given)',
            '_read_preamble() takes 1 positional argument but 2 were given'
        })

        class Bad1:
            pass

        class Bad2:
            readline = 'not callable'

        # rfile has no `readline` attribute:
        rfile = Bad1()
        self.assertEqual(sys.getrefcount(rfile), 2)
        with self.assertRaises(AttributeError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception),
            "'Bad1' object has no attribute 'readline'"
        )
        self.assertEqual(sys.getrefcount(rfile), 2)

        # `rfile.readline` is not callable:
        rfile = Bad2()
        self.assertEqual(sys.getrefcount(rfile), 2)
        with self.assertRaises(TypeError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception), 'rfile.readline is not callable')
        self.assertEqual(sys.getrefcount(rfile), 2)

        ##################################################################
        # `rfile.readline()` raises an exception, doesn't return bytes, or
        # returns too many bytes... all on the first line:

        # Exception raised inside call to `rfile.readline()`:
        rfile = DummyFile([])
        self.assertEqual(sys.getrefcount(rfile), 2)
        with self.assertRaises(IndexError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception), 'pop from empty list')
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls, [backend._MAX_LINE_SIZE])
        self.assertEqual(sys.getrefcount(rfile), 2)

        # `rfile.readline()` doesn't return bytes:
        lines = [random_line().decode()]
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        with self.assertRaises(TypeError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception),
            'rfile.readline() returned {!r}, should return {!r}'.format(str, bytes)
        )
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls, [backend._MAX_LINE_SIZE])
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )

        lines = [UserBytes(random_line())]
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        with self.assertRaises(TypeError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception),
            'rfile.readline() returned {!r}, should return {!r}'.format(UserBytes, bytes)
        )
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls, [backend._MAX_LINE_SIZE])
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )

        # `rfile.readline()` returns more than *size* bytes:
        lines = [b'D' * (backend._MAX_LINE_SIZE - 1) + b'\r\n']
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        with self.assertRaises(ValueError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception),
            'rfile.readline() returned 4097 bytes, expected at most 4096'
        )
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls, [backend._MAX_LINE_SIZE])
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )

        ##################################################################
        # `rfile.readline()` raises an exception, doesn't return bytes, or
        # returns too many bytes... all on the first *header* line:

        # Exception raised inside call to `rfile.readline()`:
        lines = [random_line()]
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        self.assertEqual(sys.getrefcount(rfile), 2)
        with self.assertRaises(IndexError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception), 'pop from empty list')
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls,
            [backend._MAX_LINE_SIZE, backend._MAX_LINE_SIZE]
        )
        self.assertEqual(sys.getrefcount(rfile), 2)

        # `rfile.readline()` doesn't return bytes:
        lines = [random_line(), random_header_line().decode()]
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        with self.assertRaises(TypeError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception),
            'rfile.readline() returned {!r}, should return {!r}'.format(str, bytes)
        )
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls,
            [backend._MAX_LINE_SIZE, backend._MAX_LINE_SIZE]
        )
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )

        lines = [random_line(), UserBytes(random_header_line())]
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        with self.assertRaises(TypeError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception),
            'rfile.readline() returned {!r}, should return {!r}'.format(UserBytes, bytes)
        )
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls,
            [backend._MAX_LINE_SIZE, backend._MAX_LINE_SIZE]
        )
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )

        # `rfile.readline()` returns more than *size* bytes:
        lines = [
            random_line(),
            b'D' * (backend._MAX_LINE_SIZE - 1) + b'\r\n',
        ]
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        with self.assertRaises(ValueError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception),
            'rfile.readline() returned 4097 bytes, expected at most 4096'
        )
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls,
            [backend._MAX_LINE_SIZE, backend._MAX_LINE_SIZE]
        )
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )

        ##################################################################
        # `rfile.readline()` raises an exception, doesn't return bytes, or
        # returns too many bytes... all on the *last* header line:

        # Exception raised inside call to `rfile.readline()`:
        lines = [random_line()]
        lines.extend(
            random_header_line() for i in range(backend._MAX_HEADER_COUNT - 1)
        )
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        self.assertEqual(sys.getrefcount(rfile), 2)
        with self.assertRaises(IndexError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception), 'pop from empty list')
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls,
            [backend._MAX_LINE_SIZE for i in range(backend._MAX_HEADER_COUNT + 1)]
        )
        self.assertEqual(sys.getrefcount(rfile), 2)

        # `rfile.readline()` doesn't return bytes:
        lines = [random_line()]
        lines.extend(
            random_header_line() for i in range(backend._MAX_HEADER_COUNT - 1)
        )
        lines.append(random_header_line().decode())
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        with self.assertRaises(TypeError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception),
            'rfile.readline() returned {!r}, should return {!r}'.format(str, bytes)
        )
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls,
            [backend._MAX_LINE_SIZE for i in range(backend._MAX_HEADER_COUNT + 1)]
        )
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )

        lines = [random_line()]
        lines.extend(
            random_header_line() for i in range(backend._MAX_HEADER_COUNT - 1)
        )
        lines.append(UserBytes(random_header_line()))
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        with self.assertRaises(TypeError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception),
            'rfile.readline() returned {!r}, should return {!r}'.format(UserBytes, bytes)
        )
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls,
            [backend._MAX_LINE_SIZE for i in range(backend._MAX_HEADER_COUNT + 1)]
        )
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )

        # `rfile.readline()` returns more than *size* bytes:
        lines = [random_line()]
        lines.extend(
            random_header_line() for i in range(backend._MAX_HEADER_COUNT - 1)
        )
        lines.append(b'D' * (backend._MAX_LINE_SIZE - 1) + b'\r\n')
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        with self.assertRaises(ValueError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception),
            'rfile.readline() returned 4097 bytes, expected at most 4096'
        )
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls,
            [backend._MAX_LINE_SIZE for i in range(backend._MAX_HEADER_COUNT + 1)]
        )
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )

        ##################################################################
        # `rfile.readline()` raises an exception, doesn't return bytes, or
        # returns too many bytes... all on the final CRLF preamble terminating
        # line:

        # Exception raised inside call to `rfile.readline()`:
        lines = [random_line()]
        lines.extend(
            random_header_line() for i in range(backend._MAX_HEADER_COUNT)
        )
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        self.assertEqual(sys.getrefcount(rfile), 2)
        with self.assertRaises(IndexError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception), 'pop from empty list')
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls,
            [backend._MAX_LINE_SIZE for i in range(backend._MAX_HEADER_COUNT + 1)]
            + [2]
        )
        self.assertEqual(sys.getrefcount(rfile), 2)

        # `rfile.readline()` doesn't return bytes:
        lines = [random_line()]
        lines.extend(
            random_header_line() for i in range(backend._MAX_HEADER_COUNT)
        )
        lines.append('\r\n')
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        with self.assertRaises(TypeError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception),
            'rfile.readline() returned {!r}, should return {!r}'.format(str, bytes)
        )
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls,
            [backend._MAX_LINE_SIZE for i in range(backend._MAX_HEADER_COUNT + 1)]
            + [2]
        )
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )

        lines = [random_line()]
        lines.extend(
            random_header_line() for i in range(backend._MAX_HEADER_COUNT)
        )
        lines.append(UserBytes(b'\r\n'))
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        with self.assertRaises(TypeError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception),
            'rfile.readline() returned {!r}, should return {!r}'.format(UserBytes, bytes)
        )
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls,
            [backend._MAX_LINE_SIZE for i in range(backend._MAX_HEADER_COUNT + 1)]
            + [2]
        )
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )

        # `rfile.readline()` returns more than *size* bytes:
        lines = [random_line()]
        lines.extend(
            random_header_line() for i in range(backend._MAX_HEADER_COUNT)
        )
        lines.append(b'D\r\n')
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        with self.assertRaises(ValueError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception),
            'rfile.readline() returned 3 bytes, expected at most 2'
        )
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls,
            [backend._MAX_LINE_SIZE for i in range(backend._MAX_HEADER_COUNT + 1)]
            + [2]
        )
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )

        ###############################
        # Back to testing first line...

        # First line is completely empty, no termination:
        lines = [b'']
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        self.assertEqual(sys.getrefcount(rfile), 2)
        with self.assertRaises(backend.EmptyPreambleError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception), 'HTTP preamble is empty')
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls, [backend._MAX_LINE_SIZE])
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )

        # First line badly terminated:
        lines = [b'hello\n']
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        self.assertEqual(sys.getrefcount(rfile), 2)
        with self.assertRaises(ValueError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception), "bad line termination: b'o\\n'")
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls, [backend._MAX_LINE_SIZE])
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )

        # First line is empty yet well terminated:
        lines = [b'\r\n']
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        self.assertEqual(sys.getrefcount(rfile), 2)
        with self.assertRaises(ValueError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception), 'first preamble line is empty')
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls, [backend._MAX_LINE_SIZE])
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )

        ###############################
        # Back to testing header lines:

        # 1st header line is completely empty, no termination:
        lines = [random_line(), b'']
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        self.assertEqual(sys.getrefcount(rfile), 2)
        with self.assertRaises(ValueError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception), "bad header line termination: b''")
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls,
            [backend._MAX_LINE_SIZE, backend._MAX_LINE_SIZE]
        )
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )

        # 1st header line is just b'\n':
        lines = [random_line(), b'\n']
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        self.assertEqual(sys.getrefcount(rfile), 2)
        with self.assertRaises(ValueError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception),
            "bad header line termination: b'\\n'"
        )
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls,
            [backend._MAX_LINE_SIZE, backend._MAX_LINE_SIZE]
        )
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )

        # Valid header but missing \r:
        lines = [random_line(), b'Content-Length: 1776\n']
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        self.assertEqual(sys.getrefcount(rfile), 2)
        with self.assertRaises(ValueError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception),
            "bad header line termination: b'6\\n'"
        )
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls,
            [backend._MAX_LINE_SIZE, backend._MAX_LINE_SIZE]
        )
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )

        # Header line plus CRLF is fewer than six bytes in length:
        for size in [1, 2, 3]:
            for p in itertools.permutations('k: v', size):
                badline = '{}\r\n'.format(''.join(p)).encode()
                self.assertTrue(3 <= len(badline) < 6)

                # 1st header line is bad:
                lines_1 = [random_line(), badline]

                # 2nd header line is bad:
                lines_2 = [random_line(), random_header_line(), badline]

                # 3rd header line is bad:
                lines_3 = [random_line(), random_header_line(), random_header_line(), badline]

                # Test 'em all:
                for lines in (lines_1, lines_2, lines_3):
                    counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
                    rfile = DummyFile(lines.copy())
                    self.assertEqual(sys.getrefcount(rfile), 2)
                    with self.assertRaises(ValueError) as cm:
                        read_preamble(rfile)
                    self.assertEqual(str(cm.exception),
                        'header line too short: {!r}'.format(badline)
                    )
                    self.assertEqual(rfile._lines, [])
                    self.assertEqual(rfile._calls,
                        [backend._MAX_LINE_SIZE for i in range(len(lines))]
                    )
                    self.assertEqual(sys.getrefcount(rfile), 2)
                    self.assertEqual(counts,
                        tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
                    )

        # Problems in parsing header line:
        for bad in BAD_HEADER_LINES:
            lines = [random_line(), bad]
            counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
            rfile = DummyFile(lines.copy())
            self.assertEqual(sys.getrefcount(rfile), 2)
            with self.assertRaises(ValueError) as cm:
                read_preamble(rfile)
            if len(bad) < 6:
                self.assertEqual(str(cm.exception),
                    'header line too short: {!r}'.format(bad)
                )
            else:
                self.assertEqual(str(cm.exception),
                    'bad header line: {!r}'.format(bad)
                )
            self.assertEqual(rfile._lines, [])
            self.assertEqual(rfile._calls,
                [backend._MAX_LINE_SIZE, backend._MAX_LINE_SIZE]
            )
            self.assertEqual(sys.getrefcount(rfile), 2)
            self.assertEqual(counts,
                tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
            )

        # Bad Content-Length:
        lines = [random_line(), b'Content-Length: 16.9\r\n']
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        self.assertEqual(sys.getrefcount(rfile), 2)
        with self.assertRaises(ValueError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception),
            "bad bytes in content-length: b'16.9'"
        )
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls,
            [backend._MAX_LINE_SIZE for i in range(2)]
        )
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )

        # Negative Content-Length:
        lines = [random_line(), b'Content-Length: -17\r\n']
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        self.assertEqual(sys.getrefcount(rfile), 2)
        with self.assertRaises(ValueError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception),
            "bad bytes in content-length: b'-17'"
        )
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls,
            [backend._MAX_LINE_SIZE for i in range(2)]
        )
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )

        # Bad Transfer-Encoding:
        lines = [random_line(), b'Transfer-Encoding: clumped\r\n']
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        self.assertEqual(sys.getrefcount(rfile), 2)
        with self.assertRaises(ValueError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception), "bad transfer-encoding: b'clumped'")
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls,
            [backend._MAX_LINE_SIZE for i in range(2)]
        )
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )

        # Duplicate header:
        lines = [
            random_line(),
            b'content-type: text/plain\r\n',
            b'Content-Type: text/plain\r\n',
        ]
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        self.assertEqual(sys.getrefcount(rfile), 2)
        with self.assertRaises(ValueError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception),
            "duplicate header: b'Content-Type: text/plain'"
        )
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls,
            [backend._MAX_LINE_SIZE for i in range(3)]
        )
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )

        # Content-Length with Transfer-Encoding:
        lines = [
            random_line(),
            b'Content-Length: 17\r\n',
            b'Transfer-Encoding: chunked\r\n',
            b'\r\n',
        ]
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        self.assertEqual(sys.getrefcount(rfile), 2)
        with self.assertRaises(ValueError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception),
            'cannot have both content-length and transfer-encoding headers'
        )
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls,
            [backend._MAX_LINE_SIZE for i in range(4)]
        )
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )

        # content-length with transfer-encoding:
        lines = [
            random_line(),
            b'content-length: 17\r\n',
            b'transfer-encoding: chunked\r\n',
            b'\r\n',
        ]
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        self.assertEqual(sys.getrefcount(rfile), 2)
        with self.assertRaises(ValueError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception),
            'cannot have both content-length and transfer-encoding headers'
        )
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls,
            [backend._MAX_LINE_SIZE for i in range(4)]
        )
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )

        # Too many headers:
        first_line = random_line()
        header_lines = tuple(
            random_header_line() for i in range(backend._MAX_HEADER_COUNT)
        )
        lines = [first_line]
        lines.extend(header_lines)
        lines.append(b'D\n')
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        self.assertEqual(sys.getrefcount(rfile), 2)
        with self.assertRaises(ValueError) as cm:
            read_preamble(rfile)
        self.assertEqual(str(cm.exception),
            'too many headers (> {!r})'.format(backend._MAX_HEADER_COUNT)
        )
        self.assertEqual(rfile._lines, [])
        calls = [
            backend._MAX_LINE_SIZE for i in range(backend._MAX_HEADER_COUNT + 1)
        ]
        calls.append(2)
        self.assertEqual(rfile._calls, calls)
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )

        # Test a number of good single values:
        for (header_line, (key, value)) in GOOD_HEADERS:
            first_line = random_line()
            lines = [first_line, header_line, b'\r\n']
            counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
            rfile = DummyFile(lines.copy())
            self.assertEqual(sys.getrefcount(rfile), 2)
            (first, headers) = read_preamble(rfile)
            self.assertEqual(sys.getrefcount(first), 2)
            self.assertEqual(sys.getrefcount(headers), 2)
            self.assertEqual(rfile._lines, [])
            self.assertEqual(rfile._calls,
                [backend._MAX_LINE_SIZE for i in range(3)]
            )
            self.assertEqual(sys.getrefcount(rfile), 2)
            self.assertEqual(counts,
                tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
            )
            self.assertIsInstance(first, str)
            self.assertEqual(first, first_line[:-2].decode('latin_1'))
            self.assertIsInstance(headers, dict)
            self.assertEqual(headers, {key: value})

        # No headers:
        first_line = random_line()
        lines = [first_line, b'\r\n']
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        self.assertEqual(sys.getrefcount(rfile), 2)
        (first, headers) = read_preamble(rfile)
        self.assertEqual(sys.getrefcount(first), 2)
        self.assertEqual(sys.getrefcount(headers), 2)
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls,
            [backend._MAX_LINE_SIZE, backend._MAX_LINE_SIZE]
        )
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )
        self.assertIsInstance(first, str)
        self.assertEqual(first, first_line[:-2].decode('latin_1'))
        self.assertIsInstance(headers, dict)
        self.assertEqual(headers, {})

        # 1 header:
        first_line = random_line()
        header_line = random_header_line()
        lines = [first_line, header_line, b'\r\n']
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        self.assertEqual(sys.getrefcount(rfile), 2)
        (first, headers) = read_preamble(rfile)
        self.assertEqual(sys.getrefcount(first), 2)
        self.assertEqual(sys.getrefcount(headers), 2)
        for kv in headers.items():
            self.assertEqual(sys.getrefcount(kv[0]), 3)
            self.assertEqual(sys.getrefcount(kv[1]), 3)
        self.assertEqual(rfile._lines, [])
        self.assertEqual(rfile._calls,
            [backend._MAX_LINE_SIZE for i in range(3)]
        )
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )
        self.assertIsInstance(first, str)
        self.assertEqual(first, first_line[:-2].decode('latin_1'))
        self.assertIsInstance(headers, dict)
        self.assertEqual(len(headers), 1)
        key = header_line.split(b': ')[0].decode('latin_1').lower()
        value = headers[key]
        self.assertIsInstance(value, str)
        self.assertEqual(value,
            header_line[:-2].split(b': ')[1].decode('latin_1')
        )

        # _MAX_HEADER_COUNT:
        first_line = random_line()
        header_lines = tuple(
            random_header_line() for i in range(backend._MAX_HEADER_COUNT)
        )
        lines = [first_line]
        lines.extend(header_lines)
        lines.append(b'\r\n')
        counts = tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        rfile = DummyFile(lines.copy())
        self.assertEqual(sys.getrefcount(rfile), 2)
        (first, headers) = read_preamble(rfile)
        self.assertEqual(sys.getrefcount(first), 2)
        self.assertEqual(sys.getrefcount(headers), 2)
        for kv in headers.items():
            self.assertEqual(sys.getrefcount(kv[0]), 3)
            self.assertEqual(sys.getrefcount(kv[1]), 3)
        self.assertEqual(rfile._lines, [])
        calls = [
            backend._MAX_LINE_SIZE for i in range(backend._MAX_HEADER_COUNT + 1)
        ]
        calls.append(2)
        self.assertEqual(rfile._calls, calls)
        self.assertEqual(sys.getrefcount(rfile), 2)
        self.assertEqual(counts,
            tuple(sys.getrefcount(lines[i]) for i in range(len(lines)))
        )
        self.assertIsInstance(first, str)
        self.assertEqual(first, first_line[:-2].decode('latin_1'))
        self.assertIsInstance(headers, dict)
        self.assertEqual(len(headers), len(header_lines))
        for line in header_lines:
            key = line.split(b': ')[0].decode('latin_1').lower()
            value = headers[key]
            self.assertIsInstance(value, str)
            self.assertEqual(value, line[:-2].split(b': ')[1].decode('latin_1'))

    def test__read_preamble_p(self):
        self.check__read_preamble(_basepy)

    def test__read_preamble_c(self):
        self.skip_if_no_c_ext()
        self.check__read_preamble(_base)

    def test_read_chunk(self):
        data = (b'D' * 7777)  # Longer than _MAX_LINE_SIZE
        small_data = (b'd' * 6666)  # Still longer than _MAX_LINE_SIZE
        termed = data + b'\r\n'
        self.assertEqual(len(termed), 7779)
        size = b'1e61\r\n'
        size_plus = b'1e61;foo=bar\r\n'

        # No CRLF terminated chunk size line:
        rfile = io.BytesIO(termed)
        with self.assertRaises(ValueError) as cm:
            base.read_chunk(rfile)
        self.assertEqual(str(cm.exception),
            "bad chunk size termination: b'DD'"
        )
        self.assertEqual(rfile.tell(), _MAX_LINE_SIZE)
        self.assertFalse(rfile.closed)

        # Size line has LF but no CR:
        rfile = io.BytesIO(b'1e61\n' + termed)
        with self.assertRaises(ValueError) as cm:
            base.read_chunk(rfile)
        self.assertEqual(str(cm.exception),
            "bad chunk size termination: b'1\\n'"
        )
        self.assertEqual(rfile.tell(), 5)
        self.assertFalse(rfile.closed)

        # Totally empty:
        rfile = io.BytesIO(b'')
        with self.assertRaises(ValueError) as cm:
            base.read_chunk(rfile)
        self.assertEqual(str(cm.exception),
            "bad chunk size termination: b''"
        )
        self.assertEqual(rfile.tell(), 0)
        self.assertFalse(rfile.closed)

        # Size line is property terminated, but empty value:
        rfile = io.BytesIO(b'\r\n' + termed)
        with self.assertRaises(ValueError) as cm:
            base.read_chunk(rfile)
        self.assertEqual(str(cm.exception),
            "invalid literal for int() with base 16: b''"
        )
        self.assertEqual(rfile.tell(), 2)
        self.assertFalse(rfile.closed)

        # Too many b';' is size line:
        rfile = io.BytesIO(b'foo;bar;baz\r\ndata\r\n')
        with self.assertRaises(ValueError) as cm:
            base.read_chunk(rfile)
        self.assertEqual(str(cm.exception),
            "bad chunk size line: b'foo;bar;baz\\r\\n'"
        )
        self.assertEqual(rfile.tell(), 13)
        self.assertEqual(rfile.read(), b'data\r\n')

        # Size isn't a hexidecimal integer:
        rfile = io.BytesIO(b'17.6\r\n' + termed)
        with self.assertRaises(ValueError) as cm:
            base.read_chunk(rfile)
        self.assertEqual(str(cm.exception),
            "invalid literal for int() with base 16: b'17.6'"
        )
        self.assertEqual(rfile.tell(), 6)
        self.assertFalse(rfile.closed)
        rfile = io.BytesIO(b'17.6;1e61=bar\r\n' + termed)
        with self.assertRaises(ValueError) as cm:
            base.read_chunk(rfile)
        self.assertEqual(str(cm.exception),
            "invalid literal for int() with base 16: b'17.6'"
        )
        self.assertEqual(rfile.tell(), 15)
        self.assertFalse(rfile.closed)

        # Size is negative:
        rfile = io.BytesIO(b'-1\r\n' + termed)
        with self.assertRaises(ValueError) as cm:
            base.read_chunk(rfile)
        self.assertEqual(str(cm.exception),
            'need 0 <= chunk_size <= {}; got -1'.format(base.MAX_CHUNK_SIZE)
        )
        self.assertEqual(rfile.tell(), 4)
        self.assertFalse(rfile.closed)
        rfile = io.BytesIO(b'-1e61;1e61=bar\r\n' + termed)
        with self.assertRaises(ValueError) as cm:
            base.read_chunk(rfile)
        self.assertEqual(str(cm.exception),
            'need 0 <= chunk_size <= {}; got -7777'.format(base.MAX_CHUNK_SIZE)
        )
        self.assertEqual(rfile.tell(), 16)
        self.assertFalse(rfile.closed)

        # Size > MAX_CHUNK_SIZE:
        line = '{:x}\r\n'.format(base.MAX_CHUNK_SIZE + 1)
        rfile = io.BytesIO(line.encode('latin_1') + data)
        with self.assertRaises(ValueError) as cm:
            base.read_chunk(rfile)
        self.assertEqual(str(cm.exception),
            'need 0 <= chunk_size <= 16777216; got 16777217'
        )
        self.assertEqual(rfile.tell(), len(line))
        self.assertFalse(rfile.closed)

        # Size > MAX_CHUNK_SIZE, with extension:
        line = '{:x};foo=bar\r\n'.format(base.MAX_CHUNK_SIZE + 1)
        rfile = io.BytesIO(line.encode('latin_1') + data)
        with self.assertRaises(ValueError) as cm:
            base.read_chunk(rfile)
        self.assertEqual(str(cm.exception),
            'need 0 <= chunk_size <= 16777216; got 16777217'
        )
        self.assertEqual(rfile.tell(), len(line))
        self.assertFalse(rfile.closed)

        # Too few b'=' in chunk extension:
        rfile = io.BytesIO(b'1e61;foo\r\ndata\r\n')
        with self.assertRaises(ValueError) as cm:
            base.read_chunk(rfile)
        self.assertEqual(str(cm.exception),
            'need more than 1 value to unpack'
        )
        self.assertEqual(rfile.tell(), 10)
        self.assertEqual(rfile.read(), b'data\r\n')

        # Too many b'=' in chunk extension:
        rfile = io.BytesIO(b'1e61;foo=bar=baz\r\ndata\r\n')
        with self.assertRaises(ValueError) as cm:
            base.read_chunk(rfile)
        self.assertEqual(str(cm.exception),
            'too many values to unpack (expected 2)'
        )
        self.assertEqual(rfile.tell(), 18)
        self.assertEqual(rfile.read(), b'data\r\n')

        # Not enough data:
        rfile = io.BytesIO(size + small_data + b'\r\n')
        with self.assertRaises(ValueError) as cm:
            base.read_chunk(rfile)
        self.assertEqual(str(cm.exception), 'underflow: 6668 < 7777')
        self.assertEqual(rfile.tell(), 6674)
        self.assertFalse(rfile.closed)

        # Data isn't properly terminated:
        rfile = io.BytesIO(size + data + b'TT\r\n')
        with self.assertRaises(ValueError) as cm:
            base.read_chunk(rfile)
        self.assertEqual(str(cm.exception), "bad chunk data termination: b'TT'")
        self.assertEqual(rfile.tell(), 7785)
        self.assertFalse(rfile.closed)

        # Test when it's all good:
        rfile = io.BytesIO(size + termed)
        self.assertEqual(base.read_chunk(rfile), (None, data))
        self.assertEqual(rfile.tell(), 7785)
        self.assertFalse(rfile.closed)

        # Test when size line has extra information:
        rfile = io.BytesIO(size_plus + termed)
        self.assertEqual(base.read_chunk(rfile), (('foo', 'bar'), data))
        self.assertEqual(rfile.tell(), 7793)
        self.assertFalse(rfile.closed)

        # Test max chunk size:
        data = os.urandom(base.MAX_CHUNK_SIZE)
        line = '{:x}\r\n'.format(len(data))
        rfile = io.BytesIO()
        rfile.write(line.encode('latin_1'))
        rfile.write(data)
        rfile.write(b'\r\n')
        rfile.seek(0)
        self.assertEqual(base.read_chunk(rfile), (None, data))
        self.assertEqual(rfile.tell(), len(line) + len(data) + 2)

        # Again, with extension:
        data = os.urandom(base.MAX_CHUNK_SIZE)
        line = '{:x};foo=bar\r\n'.format(len(data))
        rfile = io.BytesIO()
        rfile.write(line.encode('latin_1'))
        rfile.write(data)
        rfile.write(b'\r\n')
        rfile.seek(0)
        self.assertEqual(base.read_chunk(rfile), (('foo', 'bar'), data))
        self.assertEqual(rfile.tell(), len(line) + len(data) + 2)

        # Bad bytes in extension:
        linestart = b'6f0;'
        for bad in helpers.iter_bad_values(8):
            line = linestart + bad + b'\r\n'
            rfile = io.BytesIO(line)
            with self.assertRaises(ValueError) as cm:
                base.read_chunk(rfile)
            if b';' in bad:
                self.assertEqual(str(cm.exception),
                    'bad chunk size line: {!r}'.format(line)
                )
            else:
                self.assertEqual(str(cm.exception),
                    'bad bytes in chunk extension: {!r}'.format(bad)
                )
            self.assertEqual(sys.getrefcount(rfile), 2)
            self.assertEqual(rfile.tell(), 14)

    def test_write_chunk(self):
        # len(data) > MAX_CHUNK_SIZE:
        data = b'D' * (base.MAX_CHUNK_SIZE + 1)
        wfile = io.BytesIO()
        chunk = (None, data)
        with self.assertRaises(ValueError) as cm:
            base.write_chunk(wfile, chunk)
        self.assertEqual(str(cm.exception),
            'need len(data) <= 16777216; got 16777217'
        )
        self.assertEqual(wfile.getvalue(), b'')

        # len(data) > MAX_CHUNK_SIZE, but now with extension:
        wfile = io.BytesIO()
        chunk = (('foo', 'bar'), data)
        with self.assertRaises(ValueError) as cm:
            base.write_chunk(wfile, chunk)
        self.assertEqual(str(cm.exception),
            'need len(data) <= 16777216; got 16777217'
        )
        self.assertEqual(wfile.getvalue(), b'')

        # Empty data:
        wfile = io.BytesIO()
        chunk = (None, b'')
        self.assertEqual(base.write_chunk(wfile, chunk), 5)
        self.assertEqual(wfile.getvalue(), b'0\r\n\r\n')

        # Empty data plus extension:
        wfile = io.BytesIO()
        chunk = (('foo', 'bar'),  b'')
        self.assertEqual(base.write_chunk(wfile, chunk), 13)
        self.assertEqual(wfile.getvalue(), b'0;foo=bar\r\n\r\n')

        # Small data:
        wfile = io.BytesIO()
        chunk = (None, b'hello')
        self.assertEqual(base.write_chunk(wfile, chunk), 10)
        self.assertEqual(wfile.getvalue(), b'5\r\nhello\r\n')

        # Small data plus extension:
        wfile = io.BytesIO()
        chunk = (('foo', 'bar'), b'hello')
        self.assertEqual(base.write_chunk(wfile, chunk), 18)
        self.assertEqual(wfile.getvalue(), b'5;foo=bar\r\nhello\r\n')

        # Larger data:
        data = b'D' * 7777
        wfile = io.BytesIO()
        chunk = (None, data)
        self.assertEqual(base.write_chunk(wfile, chunk), 7785)
        self.assertEqual(wfile.getvalue(), b'1e61\r\n' + data + b'\r\n')

        # Larger data plus extension:
        wfile = io.BytesIO()
        chunk = (('foo', 'bar'), data)
        self.assertEqual(base.write_chunk(wfile, chunk), 7793)
        self.assertEqual(wfile.getvalue(), b'1e61;foo=bar\r\n' + data + b'\r\n')

        # Test random value round-trip with read_chunk():
        for size in range(1776):
            # No extension:
            data = os.urandom(size)
            total = size + len('{:x}'.format(size)) + 4
            fp = io.BytesIO()
            chunk = (None, data)
            self.assertEqual(base.write_chunk(fp, chunk), total)
            fp.seek(0)
            self.assertEqual(base.read_chunk(fp), chunk)

            # With extension:
            key = random_id()
            value = random_id()
            total = size + len('{:x};{}={}'.format(size, key, value)) + 4
            fp = io.BytesIO()
            chunk = ((key, value), data)
            self.assertEqual(base.write_chunk(fp, chunk), total)
            fp.seek(0)
            self.assertEqual(base.read_chunk(fp), chunk)

        # Make sure we can round-trip MAX_CHUNK_SIZE:
        size = base.MAX_CHUNK_SIZE
        data = os.urandom(size)
        total = size + len('{:x}'.format(size)) + 4
        fp = io.BytesIO()
        chunk = (None, data)
        self.assertEqual(base.write_chunk(fp, chunk), total)
        fp.seek(0)
        self.assertEqual(base.read_chunk(fp), chunk)

        # With extension:
        key = random_id()
        value = random_id()
        total = size + len('{:x};{}={}'.format(size, key, value)) + 4
        chunk = ((key, value), data)
        fp = io.BytesIO()
        self.assertEqual(base.write_chunk(fp, chunk), total)
        fp.seek(0)
        self.assertEqual(base.read_chunk(fp), chunk)


class TestBody(TestCase):
    def test_init(self):
        rfile = io.BytesIO()

        # Bad content_length type:
        with self.assertRaises(TypeError) as cm:
            base.Body(rfile, 17.0)
        self.assertEqual(str(cm.exception),
            base._TYPE_ERROR.format('content_length', int, float, 17.0)
        )
        with self.assertRaises(TypeError) as cm:
            base.Body(rfile, '17')
        self.assertEqual(str(cm.exception),
            base._TYPE_ERROR.format('content_length', int, str, '17')
        )

        # Bad content_length value:
        with self.assertRaises(ValueError) as cm:
            base.Body(rfile, -1)
        self.assertEqual(str(cm.exception),
            'content_length must be >= 0, got: -1'
        )
        with self.assertRaises(ValueError) as cm:
            base.Body(rfile, -17)
        self.assertEqual(str(cm.exception),
            'content_length must be >= 0, got: -17'
        )

        # Bad io_size type:
        with self.assertRaises(TypeError) as cm:
            base.Body(rfile, 17, '8192')
        self.assertEqual(str(cm.exception),
            base._TYPE_ERROR.format('io_size', int, str, '8192')
        )
        with self.assertRaises(TypeError) as cm:
            base.Body(rfile, 17, 8192.0)
        self.assertEqual(str(cm.exception),
            base._TYPE_ERROR.format('io_size', int, float, 8192.0)
        )

        # io_size too small:
        with self.assertRaises(ValueError) as cm:
            base.Body(rfile, 17, 2048)
        self.assertEqual(str(cm.exception),
            'need 4096 <= io_size <= {}; got 2048'.format(base.MAX_READ_SIZE)
        )
        with self.assertRaises(ValueError) as cm:
            base.Body(rfile, 17, 4095)
        self.assertEqual(str(cm.exception),
            'need 4096 <= io_size <= {}; got 4095'.format(base.MAX_READ_SIZE)
        )

        # io_size too big:
        size = base.MAX_READ_SIZE * 2
        with self.assertRaises(ValueError) as cm:
            base.Body(rfile, 17, size)
        self.assertEqual(str(cm.exception),
            'need 4096 <= io_size <= {}; got {}'.format(base.MAX_READ_SIZE, size)
        )
        size = base.MAX_READ_SIZE + 1
        with self.assertRaises(ValueError) as cm:
            base.Body(rfile, 17, size)
        self.assertEqual(str(cm.exception),
            'need 4096 <= io_size <= {}; got {}'.format(base.MAX_READ_SIZE, size)
        )

        # io_size not a power of 2:
        with self.assertRaises(ValueError) as cm:
            base.Body(rfile, 17, 40960)
        self.assertEqual(str(cm.exception),
            'io_size must be a power of 2; got 40960'
        )
        # io_size not a power of 2:
        with self.assertRaises(ValueError) as cm:
            base.Body(rfile, 17, 4097)
        self.assertEqual(str(cm.exception),
            'io_size must be a power of 2; got 4097'
        )

        # All good:
        body = base.Body(rfile, 17)
        self.assertIs(body.chunked, False)
        self.assertIs(body.__class__.chunked, False)
        self.assertIs(body.rfile, rfile)
        self.assertEqual(body.content_length, 17)
        self.assertIs(body.io_size, base.IO_SIZE)
        self.assertIs(body.closed, False)
        self.assertEqual(body._remaining, 17)
        self.assertEqual(repr(body), 'Body(<rfile>, 17)')

        # Now override io_size with a number of good values:
        for size in (4096, 8192, 1048576, base.MAX_READ_SIZE):
            body = base.Body(rfile, 17, size)
            self.assertIs(body.io_size, size)
            body = base.Body(rfile, 17, io_size=size)
            self.assertIs(body.io_size, size)

    def test_len(self):
        for content_length in (0, 17, 27, 37):
            body = base.Body(io.BytesIO(), content_length)
        self.assertEqual(len(body), content_length)

    def test_read(self):
        data = os.urandom(1776)
        rfile = io.BytesIO(data)
        body = base.Body(rfile, len(data))

        # body.closed is True:
        body.closed = True
        with self.assertRaises(ValueError) as cm:
            body.read()
        self.assertEqual(str(cm.exception), 'Body.closed, already consumed')
        self.assertIs(body.chunked, False)
        self.assertIs(body.closed, True)
        self.assertEqual(rfile.tell(), 0)
        self.assertEqual(body.content_length, 1776)
        self.assertEqual(body._remaining, 1776)

        # Bad size type:
        body.closed = False
        with self.assertRaises(TypeError) as cm:
            body.read(18.0)
        self.assertEqual(str(cm.exception),
            base._TYPE_ERROR.format('size', int, float, 18.0)
        )
        self.assertIs(body.chunked, False)
        self.assertIs(body.closed, False)
        self.assertEqual(rfile.tell(), 0)
        self.assertEqual(body.content_length, 1776)
        self.assertEqual(body._remaining, 1776)
        with self.assertRaises(TypeError) as cm:
            body.read('18')
        self.assertEqual(str(cm.exception),
            base._TYPE_ERROR.format('size', int, str, '18')
        )
        self.assertIs(body.chunked, False)
        self.assertIs(body.closed, False)
        self.assertEqual(rfile.tell(), 0)
        self.assertEqual(body.content_length, 1776)
        self.assertEqual(body._remaining, 1776)

        # Bad size value:
        with self.assertRaises(ValueError) as cm:
            body.read(-1)
        self.assertEqual(str(cm.exception), 'size must be >= 0; got -1')
        self.assertIs(body.chunked, False)
        self.assertIs(body.closed, False)
        self.assertEqual(rfile.tell(), 0)
        self.assertEqual(body.content_length, 1776)
        self.assertEqual(body._remaining, 1776)
        with self.assertRaises(ValueError) as cm:
            body.read(-18)
        self.assertEqual(str(cm.exception), 'size must be >= 0; got -18')
        self.assertIs(body.chunked, False)
        self.assertIs(body.closed, False)
        self.assertEqual(rfile.tell(), 0)
        self.assertEqual(body.content_length, 1776)
        self.assertEqual(body._remaining, 1776)

        # Now read it all at once:
        self.assertEqual(body.read(), data)
        self.assertIs(body.chunked, False)
        self.assertIs(body.closed, True)
        self.assertEqual(rfile.tell(), 1776)
        self.assertEqual(body.content_length, 1776)
        self.assertEqual(body._remaining, 0)
        with self.assertRaises(ValueError) as cm:
            body.read()
        self.assertEqual(str(cm.exception), 'Body.closed, already consumed')

        # Read it again, this time in parts:
        rfile = io.BytesIO(data)
        body = base.Body(rfile, 1776)
        self.assertEqual(body.read(17), data[0:17])
        self.assertIs(body.chunked, False)
        self.assertIs(body.closed, False)
        self.assertEqual(rfile.tell(), 17)
        self.assertEqual(body.content_length, 1776)
        self.assertEqual(body._remaining, 1759)

        self.assertEqual(body.read(18), data[17:35])
        self.assertIs(body.chunked, False)
        self.assertIs(body.closed, False)
        self.assertEqual(rfile.tell(), 35)
        self.assertEqual(body.content_length, 1776)
        self.assertEqual(body._remaining, 1741)

        self.assertEqual(body.read(1741), data[35:])
        self.assertIs(body.chunked, False)
        self.assertIs(body.closed, False)
        self.assertEqual(rfile.tell(), 1776)
        self.assertEqual(body.content_length, 1776)
        self.assertEqual(body._remaining, 0)

        self.assertEqual(body.read(1776), b'')
        self.assertIs(body.chunked, False)
        self.assertIs(body.closed, True)
        self.assertEqual(rfile.tell(), 1776)
        self.assertEqual(body.content_length, 1776)
        self.assertEqual(body._remaining, 0)

        with self.assertRaises(ValueError) as cm:
            body.read(17)
        self.assertEqual(str(cm.exception), 'Body.closed, already consumed')

        # ValueError (underflow) when trying to read all:
        rfile = io.BytesIO(data)
        body = base.Body(rfile, 1800)
        with self.assertRaises(ValueError) as cm:
            body.read()
        self.assertEqual(str(cm.exception), 'underflow: 1776 < 1800')
        self.assertIs(body.closed, False)
        self.assertIs(rfile.closed, True)

        # ValueError (underflow) error when read in parts:
        data = os.urandom(35)
        rfile = io.BytesIO(data)
        body = base.Body(rfile, 37)
        self.assertEqual(body.read(18), data[:18])
        self.assertIs(body.chunked, False)
        self.assertIs(body.closed, False)
        self.assertEqual(rfile.tell(), 18)
        self.assertEqual(body.content_length, 37)
        self.assertEqual(body._remaining, 19)
        with self.assertRaises(ValueError) as cm:
            body.read(19)
        self.assertEqual(str(cm.exception), 'underflow: 17 < 19')
        self.assertIs(body.closed, False)
        self.assertIs(rfile.closed, True)

        # Test with empty body:
        rfile = io.BytesIO(os.urandom(21))
        body = base.Body(rfile, 0)
        self.assertEqual(body.read(17), b'')
        self.assertIs(body.chunked, False)
        self.assertIs(body.closed, True)
        self.assertEqual(rfile.tell(), 0)
        self.assertEqual(body.content_length, 0)
        self.assertEqual(body._remaining, 0)
        with self.assertRaises(ValueError) as cm:
            body.read(17)
        self.assertEqual(str(cm.exception), 'Body.closed, already consumed')

        # Test with random chunks:
        for i in range(25):
            chunks = random_chunks()
            assert chunks[-1] == b''
            data = b''.join(chunks)
            trailer = os.urandom(17)
            rfile = io.BytesIO(data + trailer)
            body = base.Body(rfile, len(data))
            for chunk in chunks:
                self.assertEqual(body.read(len(chunk)), chunk)
            self.assertIs(body.chunked, False)
            self.assertIs(body.closed, True)
            self.assertEqual(rfile.tell(), len(data))
            self.assertEqual(body.content_length, len(data))
            self.assertEqual(body._remaining, 0)
            with self.assertRaises(ValueError) as cm:
                body.read(17)
            self.assertEqual(str(cm.exception), 'Body.closed, already consumed')
            self.assertEqual(rfile.read(), trailer)

        # Test when read size > MAX_READ_SIZE:
        rfile = io.BytesIO()
        content_length = base.MAX_READ_SIZE + 1
        body = base.Body(rfile, content_length)
        self.assertIs(body.content_length, content_length)
        with self.assertRaises(ValueError) as cm:
            body.read()
        self.assertEqual(str(cm.exception),
            'max read size exceeded: {} > {}'.format(
                content_length, base.MAX_READ_SIZE
            )
        )

    def test_iter(self):
        data = os.urandom(1776)

        # content_length=0
        rfile = io.BytesIO(data)
        body = base.Body(rfile, 0)
        self.assertEqual(list(body), [])
        self.assertEqual(body._remaining, 0)
        self.assertIs(body.closed, True)
        with self.assertRaises(ValueError) as cm:
            list(body)
        self.assertEqual(str(cm.exception), 'Body.closed, already consumed')
        self.assertEqual(rfile.tell(), 0)
        self.assertEqual(rfile.read(), data)

        # content_length=69
        rfile = io.BytesIO(data)
        body = base.Body(rfile, 69)
        self.assertEqual(list(body), [data[:69]])
        self.assertEqual(body._remaining, 0)
        self.assertIs(body.closed, True)
        with self.assertRaises(ValueError) as cm:
            list(body)
        self.assertEqual(str(cm.exception), 'Body.closed, already consumed')
        self.assertEqual(rfile.tell(), 69)
        self.assertEqual(rfile.read(), data[69:])

        # content_length=1776
        rfile = io.BytesIO(data)
        body = base.Body(rfile, 1776)
        self.assertEqual(list(body), [data])
        self.assertEqual(body._remaining, 0)
        self.assertIs(body.closed, True)
        with self.assertRaises(ValueError) as cm:
            list(body)
        self.assertEqual(str(cm.exception), 'Body.closed, already consumed')
        self.assertEqual(rfile.tell(), 1776)
        self.assertEqual(rfile.read(), b'')

        # content_length=1777
        rfile = io.BytesIO(data)
        body = base.Body(rfile, 1777)
        with self.assertRaises(ValueError) as cm:
            list(body)
        self.assertEqual(str(cm.exception), 'underflow: 1776 < 1777')
        self.assertIs(body.closed, False)
        self.assertIs(rfile.closed, True)

        # Make sure data is read in IO_SIZE chunks:
        data1 = os.urandom(base.IO_SIZE)
        data2 = os.urandom(base.IO_SIZE)
        length = base.IO_SIZE * 2
        rfile = io.BytesIO(data1 + data2)
        body = base.Body(rfile, length)
        self.assertEqual(list(body), [data1, data2])
        self.assertEqual(body._remaining, 0)
        self.assertIs(body.closed, True)
        with self.assertRaises(ValueError) as cm:
            list(body)
        self.assertEqual(str(cm.exception), 'Body.closed, already consumed')
        self.assertEqual(rfile.tell(), length)
        self.assertEqual(rfile.read(), b'')

        # Again, with smaller final chunk:
        length = base.IO_SIZE * 2 + len(data)
        rfile = io.BytesIO(data1 + data2 + data)
        body = base.Body(rfile, length)
        self.assertEqual(list(body), [data1, data2, data])
        self.assertEqual(body._remaining, 0)
        self.assertIs(body.closed, True)
        with self.assertRaises(ValueError) as cm:
            list(body)
        self.assertEqual(str(cm.exception), 'Body.closed, already consumed')
        self.assertEqual(rfile.tell(), length)
        self.assertEqual(rfile.read(), b'')

        # Again, with length 1 byte less than available:
        length = base.IO_SIZE * 2 + len(data) - 1
        rfile = io.BytesIO(data1 + data2 + data)
        body = base.Body(rfile, length)
        self.assertEqual(list(body), [data1, data2, data[:-1]])
        self.assertEqual(body._remaining, 0)
        self.assertIs(body.closed, True)
        with self.assertRaises(ValueError) as cm:
            list(body)
        self.assertEqual(str(cm.exception), 'Body.closed, already consumed')
        self.assertEqual(rfile.tell(), length)
        self.assertEqual(rfile.read(), data[-1:])

        # Again, with length 1 byte *more* than available:
        length = base.IO_SIZE * 2 + len(data) + 1
        rfile = io.BytesIO(data1 + data2 + data)
        body = base.Body(rfile, length)
        with self.assertRaises(ValueError) as cm:
            list(body)
        self.assertEqual(str(cm.exception), 'underflow: 1776 < 1777')
        self.assertIs(body.closed, False)
        self.assertIs(rfile.closed, True)


class TestChunkedBody(TestCase):
    def test_init(self):
        # All good:
        rfile = io.BytesIO()
        body = base.ChunkedBody(rfile)
        self.assertIs(body.chunked, True)
        self.assertIs(body.__class__.chunked, True)
        self.assertIs(body.rfile, rfile)
        self.assertIs(body.closed, False)
        self.assertEqual(repr(body), 'ChunkedBody(<rfile>)')

    def test_readchunk(self):
        chunks = random_chunks()
        self.assertEqual(chunks[-1], b'')
        rfile = io.BytesIO()
        total = sum(base.write_chunk(rfile, (None, data)) for data in chunks)
        self.assertEqual(rfile.tell(), total)
        extra = os.urandom(3469)
        rfile.write(extra)
        rfile.seek(0)

        # Test when closed:
        body = base.ChunkedBody(rfile)
        body.closed = True
        with self.assertRaises(ValueError) as cm:
            body.readchunk()
        self.assertEqual(str(cm.exception),
            'ChunkedBody.closed, already consumed'
        )
        self.assertEqual(rfile.tell(), 0)
        self.assertIs(rfile.closed, False)

        # Test when all good:
        body = base.ChunkedBody(rfile)
        for data in chunks:
            self.assertEqual(body.readchunk(), (None, data))
        self.assertIs(body.closed, True)
        self.assertIs(rfile.closed, False)
        self.assertEqual(rfile.tell(), total)
        with self.assertRaises(ValueError) as cm:
            body.readchunk()
        self.assertEqual(str(cm.exception),
            'ChunkedBody.closed, already consumed'
        )
        self.assertEqual(rfile.read(), extra)

        # Test when read_chunk() raises an exception, which should close the
        # rfile, but not close the body:
        rfile = io.BytesIO(b'17.6\r\n' + extra)
        body = base.ChunkedBody(rfile)
        with self.assertRaises(ValueError) as cm:
            body.readchunk()
        self.assertEqual(str(cm.exception),
            "invalid literal for int() with base 16: b'17.6'"
        )
        self.assertIs(body.closed, False)
        self.assertIs(rfile.closed, True)

    def test_iter(self):
        chunks = random_chunks()
        self.assertEqual(chunks[-1], b'')
        rfile = io.BytesIO()
        total = sum(base.write_chunk(rfile, (None, data)) for data in chunks)
        self.assertEqual(rfile.tell(), total)
        extra = os.urandom(3469)
        rfile.write(extra)
        rfile.seek(0)

        # Test when closed:
        body = base.ChunkedBody(rfile)
        body.closed = True
        with self.assertRaises(ValueError) as cm:
            list(body)
        self.assertEqual(str(cm.exception),
            'ChunkedBody.closed, already consumed'
        )
        self.assertEqual(rfile.tell(), 0)
        self.assertIs(rfile.closed, False)

        # Test when all good:
        body = base.ChunkedBody(rfile)
        self.assertEqual(list(body), [(None, data) for data in chunks])
        self.assertIs(body.closed, True)
        self.assertIs(rfile.closed, False)
        self.assertEqual(rfile.tell(), total)
        with self.assertRaises(ValueError) as cm:
            list(body)
        self.assertEqual(str(cm.exception),
            'ChunkedBody.closed, already consumed'
        )
        self.assertEqual(rfile.read(), extra)

        # Test when read_chunk() raises an exception, which should close the
        # rfile, but not close the body:
        rfile = io.BytesIO(b'17.6\r\n' + extra)
        body = base.ChunkedBody(rfile)
        with self.assertRaises(ValueError) as cm:
            list(body)
        self.assertEqual(str(cm.exception),
            "invalid literal for int() with base 16: b'17.6'"
        )
        self.assertIs(body.closed, False)
        self.assertIs(rfile.closed, True)

    def test_read(self):
        # Total read size too large:
        chunks = [
            (None, b'A' * base.MAX_READ_SIZE),
            (None, b'B'),
            (None, b''),
        ]
        rfile = io.BytesIO()
        for chunk in chunks:
            base.write_chunk(rfile, chunk)
        rfile.seek(0)
        body = base.ChunkedBody(rfile)
        with self.assertRaises(ValueError) as cm:
            body.read()
        self.assertEqual(str(cm.exception),
            'max read size exceeded: {:d} > {:d}'.format(
                base.MAX_READ_SIZE + 1, base.MAX_READ_SIZE
            )
        )

        # Total read size too large:
        size = base.MAX_READ_SIZE // 8
        chunks = [
            (None, bytes([i]) * size) for i in b'ABCDEFGH'
        ]
        assert len(chunks) == 8
        chunks.extend([(None, b'I'), (None, b'')])
        rfile = io.BytesIO()
        for chunk in chunks:
            base.write_chunk(rfile, chunk)
        rfile.seek(0)
        body = base.ChunkedBody(rfile)
        with self.assertRaises(ValueError) as cm:
            body.read()
        self.assertEqual(str(cm.exception),
            'max read size exceeded: {:d} > {:d}'.format(
                base.MAX_READ_SIZE + 1, base.MAX_READ_SIZE
            )
        )

        # A chunk is larger than MAX_CHUNK_SIZE:
        pretent_max_size = base.MAX_CHUNK_SIZE + 1
        chunks = [
            (None, b'A'),
            (None, b'B' * pretent_max_size),
            (None, b''),
        ]
        rfile = io.BytesIO()
        for chunk in chunks:
            base.write_chunk(rfile, chunk, max_size=pretent_max_size)
        rfile.seek(0)
        body = base.ChunkedBody(rfile)
        with self.assertRaises(ValueError) as cm:
            body.read()
        self.assertEqual(str(cm.exception),
            'need 0 <= chunk_size <= {}; got {}'.format(
                base.MAX_CHUNK_SIZE, base.MAX_CHUNK_SIZE + 1
            )
        )


class TestBodyIter(TestCase):
    def test_init(self):
        # Good source with bad content_length type:
        with self.assertRaises(TypeError) as cm:
            base.BodyIter([], 17.0)
        self.assertEqual(str(cm.exception),
            base._TYPE_ERROR.format('content_length', int, float, 17.0)
        )
        with self.assertRaises(TypeError) as cm:
            base.BodyIter([], '17')
        self.assertEqual(str(cm.exception),
            base._TYPE_ERROR.format('content_length', int, str, '17')
        )

        # Good source with bad content_length value:
        with self.assertRaises(ValueError) as cm:
            base.BodyIter([], -1)
        self.assertEqual(str(cm.exception),
            'content_length must be >= 0, got: -1'
        )
        with self.assertRaises(ValueError) as cm:
            base.BodyIter([], -17)
        self.assertEqual(str(cm.exception),
            'content_length must be >= 0, got: -17'
        )

        # All good:
        source = []
        body = base.BodyIter(source, 17)
        self.assertIs(body.chunked, False)
        self.assertIs(body.__class__.chunked, False)
        self.assertIs(body.source, source)
        self.assertEqual(body.content_length, 17)
        self.assertIs(body.closed, False)
        self.assertIs(body._started, False)

    def test_len(self):
        for content_length in (0, 17, 27, 37):
            body = base.BodyIter([], content_length)
        self.assertEqual(len(body), content_length)

    def test_write_to(self):
        source = (b'hello', b'naughty', b'nurse')

        # Test when closed:
        body = base.BodyIter(source, 17)
        body.closed = True
        wfile = DummyWriter()
        with self.assertRaises(ValueError) as cm:
            body.write_to(wfile)
        self.assertEqual(str(cm.exception), 'BodyIter.closed, already consumed')
        self.assertEqual(wfile._calls, [])

        # Test when _started:
        body = base.BodyIter(source, 17)
        body._started = True
        wfile = DummyWriter()
        with self.assertRaises(ValueError) as cm:
            body.write_to(wfile)
        self.assertEqual(str(cm.exception), 'BodyIter._started')
        self.assertIs(body.closed, False)
        self.assertEqual(wfile._calls, [])

        # Should be closed after calling write_to():
        body = base.BodyIter(source, 17)
        wfile = DummyWriter()
        self.assertEqual(body.write_to(wfile), 17)
        self.assertIs(body._started, True)
        self.assertIs(body.closed, True)
        self.assertEqual(wfile._calls, [
            ('write', b'hello'),
            ('write', b'naughty'),
            ('write', b'nurse'),
            'flush',
        ])
        with self.assertRaises(ValueError) as cm:
            body.write_to(wfile)
        self.assertEqual(str(cm.exception), 'BodyIter.closed, already consumed')

        # ValueError should be raised at first item that pushing total above
        # content_length:
        body = base.BodyIter(source, 4)
        wfile = DummyWriter()
        with self.assertRaises(ValueError) as cm:
            body.write_to(wfile)
        self.assertEqual(str(cm.exception), 'overflow: 5 > 4')
        self.assertIs(body._started, True)
        self.assertIs(body.closed, False)
        self.assertEqual(wfile._calls, [])

        body = base.BodyIter(source, 5)
        wfile = DummyWriter()
        with self.assertRaises(ValueError) as cm:
            body.write_to(wfile)
        self.assertEqual(str(cm.exception), 'overflow: 12 > 5')
        self.assertIs(body._started, True)
        self.assertIs(body.closed, False)
        self.assertEqual(wfile._calls, [('write', b'hello')])

        body = base.BodyIter(source, 12)
        wfile = DummyWriter()
        with self.assertRaises(ValueError) as cm:
            body.write_to(wfile)
        self.assertEqual(str(cm.exception), 'overflow: 17 > 12')
        self.assertIs(body._started, True)
        self.assertIs(body.closed, False)
        self.assertEqual(wfile._calls,
            [('write', b'hello'), ('write', b'naughty')]
        )

        body = base.BodyIter(source, 16)
        wfile = DummyWriter()
        with self.assertRaises(ValueError) as cm:
            body.write_to(wfile)
        self.assertEqual(str(cm.exception), 'overflow: 17 > 16')
        self.assertIs(body._started, True)
        self.assertIs(body.closed, False)
        self.assertEqual(wfile._calls,
            [('write', b'hello'), ('write', b'naughty')]
        )

        # ValueError for underflow should only be raised after all items have
        # been yielded:
        body = base.BodyIter(source, 18)
        wfile = DummyWriter()
        with self.assertRaises(ValueError) as cm:
            body.write_to(wfile)
        self.assertEqual(str(cm.exception), 'underflow: 17 < 18')
        self.assertIs(body._started, True)
        self.assertIs(body.closed, False)
        self.assertEqual(wfile._calls,
            [('write', b'hello'), ('write', b'naughty'), ('write', b'nurse')]
        )

        # Empty data items are fine:
        source = (b'', b'hello', b'', b'naughty', b'', b'nurse', b'')
        body = base.BodyIter(source, 17)
        wfile = DummyWriter()
        self.assertEqual(body.write_to(wfile), 17)
        expected = [('write', data) for data in source]
        expected.append('flush')
        self.assertEqual(wfile._calls, expected)
        self.assertIs(body._started, True)
        self.assertIs(body.closed, True)
        with self.assertRaises(ValueError) as cm:
            body.write_to(wfile)
        self.assertEqual(str(cm.exception), 'BodyIter.closed, already consumed')

        # Test with random data of varying sizes:
        source = [os.urandom(i) for i in range(50)]
        content_length = sum(range(50))
        body = base.BodyIter(source, content_length)
        wfile = DummyWriter()
        self.assertEqual(body.write_to(wfile), content_length)
        expected = [('write', data) for data in source]
        expected.append('flush')
        self.assertEqual(wfile._calls, expected)
        self.assertIs(body._started, True)
        self.assertIs(body.closed, True)
        with self.assertRaises(ValueError) as cm:
            body.write_to(wfile)
        self.assertEqual(str(cm.exception), 'BodyIter.closed, already consumed')


class TestChunkedBodyIter(TestCase):
    def test_init(self):
        source = []
        body = base.ChunkedBodyIter(source)
        self.assertIs(body.chunked, True)
        self.assertIs(body.__class__.chunked, True)
        self.assertIs(body.source, source)
        self.assertIs(body.closed, False)
        self.assertIs(body._started, False)

    def test_write_to(self):
        source = (
            (None, b'hello'),
            (None, b'naughty'),
            (None, b'nurse'),
            (None, b''),
        )

        # Test when closed:
        body = base.ChunkedBodyIter(source)
        body.closed = True
        wfile = DummyWriter()
        with self.assertRaises(ValueError) as cm:
            body.write_to(wfile)
        self.assertEqual(str(cm.exception),
            'ChunkedBodyIter.closed, already consumed'
        )
        self.assertEqual(wfile._calls, [])

        # Test when _started:
        body = base.ChunkedBodyIter(source)
        body._started = True
        wfile = DummyWriter()
        with self.assertRaises(ValueError) as cm:
            body.write_to(wfile)
        self.assertEqual(str(cm.exception), 'ChunkedBodyIter._started')
        self.assertEqual(wfile._calls, [])

        # Should close after one call:
        body = base.ChunkedBodyIter(source)
        wfile = DummyWriter()
        self.assertEqual(body.write_to(wfile), 37)
        self.assertEqual(wfile._calls, ['flush',
            ('write', b'5\r\nhello\r\n'), 'flush',
            ('write', b'7\r\nnaughty\r\n'), 'flush',
            ('write', b'5\r\nnurse\r\n'), 'flush',
            ('write', b'0\r\n\r\n'), 'flush',
        ])
        self.assertIs(body._started, True)
        self.assertIs(body.closed, True)
        with self.assertRaises(ValueError) as cm:
            body.write_to(wfile)
        self.assertEqual(str(cm.exception),
            'ChunkedBodyIter.closed, already consumed'
        )

        # Should raise a ValueError on an empty source:
        body = base.ChunkedBodyIter([])
        wfile = DummyWriter()
        with self.assertRaises(ValueError) as cm:
            body.write_to(wfile)
        self.assertEqual(str(cm.exception), 'final chunk data was not empty')
        self.assertIs(body._started, True)
        self.assertIs(body.closed, False)
        self.assertEqual(wfile._calls, ['flush'])

        # Should raise ValueError if final chunk isn't empty:
        source = (
            (None, b'hello'),
            (None, b'naughty'),
            (None, b'nurse'),
        )
        body = base.ChunkedBodyIter(source)
        wfile = DummyWriter()
        with self.assertRaises(ValueError) as cm:
            body.write_to(wfile)
        self.assertEqual(str(cm.exception), 'final chunk data was not empty')
        self.assertIs(body._started, True)
        self.assertIs(body.closed, False)
        self.assertEqual(wfile._calls, ['flush',
            ('write', b'5\r\nhello\r\n'), 'flush',
            ('write', b'7\r\nnaughty\r\n'), 'flush',
            ('write', b'5\r\nnurse\r\n'), 'flush',
        ])

        # Should raise a ValueError if empty chunk is followed by non-empty:
        source = (
            (None, b'hello'),
            (None, b'naughty'),
            (None, b''),
            (None, b'nurse'),
            (None, b''),
        )
        body = base.ChunkedBodyIter(source)
        wfile = DummyWriter()
        with self.assertRaises(ValueError) as cm:
            body.write_to(wfile)
        self.assertEqual(str(cm.exception), 'non-empty chunk data after empty')
        self.assertIs(body._started, True)
        self.assertIs(body.closed, False)
        self.assertEqual(wfile._calls, ['flush',
            ('write', b'5\r\nhello\r\n'), 'flush',
            ('write', b'7\r\nnaughty\r\n'), 'flush',
            ('write', b'0\r\n\r\n'), 'flush',
        ])

        # Test with random data of varying sizes:
        source = [(None, os.urandom(size)) for size in range(1, 51)]
        random.shuffle(source)
        source.append((None, b''))
        body = base.ChunkedBodyIter(tuple(source))
        wfile = DummyWriter()
        self.assertEqual(body.write_to(wfile), 1565)
        self.assertIs(body._started, True)
        self.assertIs(body.closed, True)
        expected = ['flush']
        for chunk in source:
            expected.extend(
                [('write', base._encode_chunk(chunk)), 'flush']
            )
        self.assertEqual(wfile._calls, expected)




# degu: an embedded HTTP server and client library
# Copyright (C) 2014 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Unit tests for the `degu.server` module`
"""

from unittest import TestCase
import os
import io
import socket
import ssl

from .helpers import DummySocket, FuzzTestCase
from degu.base import _TYPE_ERROR
from degu.sslhelpers import random_id
from degu.misc import TempPKI
from degu import base, client


# Good type and value permutations for the CLient *address*:
GOOD_ADDRESSES = (

    # 2-tuple for AF_INET or AF_INET6:
    ('www.wikipedia.org', 80),
    ('208.80.154.224', 80),
    ('2620:0:861:ed1a::1', 80),

    # 4-tuple for AF_INET6:
    ('2620:0:861:ed1a::1', 80, 0, 0),
    ('fe80::e8b:fdff:fe75:402c', 80, 0, 3),  # Link-local

    # str for AF_UNIX:
    '/tmp/my.socket',

    # bytes for AF_UNIX (Linux abstract name):
    b'\x0000022',
)


# Some bad address permutations:
BAD_TUPLE_ADDRESSES = (
    ('::1',),
    ('127.0.0.1',),
    ('::1', 5678, 0),
    ('127.0.0.1', 5678, 0),
    ('::1', 5678, 0, 0, 0),
    ('127.0.0.1', 5678, 0, 0, 0),
)


class TestNamedTuples(TestCase):
    def test_Response(self):
        tup = client.Response('da status', 'da reason', 'da headers', 'da body')
        self.assertIsInstance(tup, tuple)
        self.assertEqual(tup.status, 'da status')
        self.assertEqual(tup.reason, 'da reason')
        self.assertEqual(tup.headers, 'da headers')
        self.assertEqual(tup.body, 'da body')


class TestClosedConnectionError(TestCase):
    def test_init(self):
        conn = random_id()
        exc = client.ClosedConnectionError(conn)
        self.assertIs(exc.conn, conn)
        self.assertEqual(str(exc),
            'cannot use request() when connection is closed: {!r}'.format(conn)
        )


class TestUnconsumedResponseError(TestCase):
    def test_init(self):
        body = random_id()
        exc = client.UnconsumedResponseError(body)
        self.assertIs(exc.body, body)
        self.assertEqual(str(exc),
            'previous response body not consumed: {!r}'.format(body)
        )


class FuzzTestFunctions(FuzzTestCase):
    def test__read_response(self):
        for method in ('GET', 'HEAD', 'DELETE', 'PUT', 'POST'):
            self.fuzz(client._read_response, base.bodies, method)


class TestFunctions(TestCase):
    def test_build_client_sslctx(self):
        # Bad sslconfig type:
        with self.assertRaises(TypeError) as cm:
            client.build_client_sslctx('bad')
        self.assertEqual(str(cm.exception),
            _TYPE_ERROR.format('sslconfig', dict, str, 'bad')
        )
 
        # The remaining test both build_client_sslctx() directly, and the
        # pass-through from _validate_client_sslctx():
        client_sslctx_funcs = (
            client.build_client_sslctx,
            client._validate_client_sslctx,
        )

        # Bad sslconfig['check_hostname'] type:
        for func in client_sslctx_funcs:
            with self.assertRaises(TypeError) as cm:
                func({'check_hostname': 0})
            self.assertEqual(str(cm.exception),
                _TYPE_ERROR.format("sslconfig['check_hostname']", bool, int, 0)
            )

        # sslconfig['key_file'] without sslconfig['cert_file']:
        for func in client_sslctx_funcs:
            with self.assertRaises(ValueError) as cm:
                func({'key_file': '/my/client.key'})
            self.assertEqual(str(cm.exception), 
                "sslconfig['key_file'] provided without sslconfig['cert_file']"
            )

        # Non absulute, non normalized paths:
        good = {
            'ca_file': '/my/sever.ca',
            'ca_path': '/my/sever.ca.dir',
            'cert_file': '/my/client.cert',
            'key_file': '/my/client.key',
        }
        for key in good.keys():
            # Relative path:
            for func in client_sslctx_funcs:
                bad = good.copy()
                value = 'relative/path'
                bad[key] = value
                with self.assertRaises(ValueError) as cm:
                    func(bad)
                self.assertEqual(str(cm.exception),
                    'sslconfig[{!r}] is not an absulute, normalized path: {!r}'.format(key, value)
                )

            # Non-normalized path with directory traversal:
            for func in client_sslctx_funcs:
                bad = good.copy()
                value = '/my/../secret/path'
                bad[key] = value
                with self.assertRaises(ValueError) as cm:
                    func(bad)
                self.assertEqual(str(cm.exception),
                    'sslconfig[{!r}] is not an absulute, normalized path: {!r}'.format(key, value)
                )

            # Non-normalized path with trailing slash:
            for func in client_sslctx_funcs:
                bad = good.copy()
                value = '/sorry/very/strict/'
                bad[key] = value
                with self.assertRaises(ValueError) as cm:
                    func(bad)
                self.assertEqual(str(cm.exception),
                    'sslconfig[{!r}] is not an absulute, normalized path: {!r}'.format(key, value)
                )

        # Empty sslconfig, will verify against system-wide CAs, and
        # check_hostname should default to True:
        for func in client_sslctx_funcs:
            sslctx = func({})
            self.assertIsInstance(sslctx, ssl.SSLContext)
            self.assertEqual(sslctx.protocol, ssl.PROTOCOL_TLSv1_2)
            self.assertEqual(sslctx.verify_mode, ssl.CERT_REQUIRED)
            self.assertIs(sslctx.check_hostname, True)

        # We don't not allow check_hostname to be False when verifying against
        # the system-wide CAs:
        for func in client_sslctx_funcs:
            with self.assertRaises(ValueError) as cm:
                func({'check_hostname': False})
            self.assertEqual(str(cm.exception),
                'check_hostname must be True when using default verify paths'
            )

        # Should work fine when explicitly providing {'check_hostname': True}:
        for func in client_sslctx_funcs:
            sslctx = func({'check_hostname': True})
            self.assertIsInstance(sslctx, ssl.SSLContext)
            self.assertEqual(sslctx.protocol, ssl.PROTOCOL_TLSv1_2)
            self.assertEqual(sslctx.verify_mode, ssl.CERT_REQUIRED)
            self.assertIs(sslctx.check_hostname, True)

        # Authenticated client sslconfig:
        pki = TempPKI()
        sslconfig = pki.client_sslconfig
        self.assertEqual(set(sslconfig),
            {'ca_file', 'cert_file', 'key_file', 'check_hostname'}
        )
        self.assertIs(sslconfig['check_hostname'], False)
        for func in client_sslctx_funcs:
            sslctx = func(sslconfig)
            self.assertIsInstance(sslctx, ssl.SSLContext)
            self.assertEqual(sslctx.protocol, ssl.PROTOCOL_TLSv1_2)
            self.assertEqual(sslctx.verify_mode, ssl.CERT_REQUIRED)
            self.assertIs(sslctx.check_hostname, False)

        # check_hostname should default to True:
        del sslconfig['check_hostname']
        for func in client_sslctx_funcs:
            sslctx = func(sslconfig)
            self.assertIsInstance(sslctx, ssl.SSLContext)
            self.assertEqual(sslctx.protocol, ssl.PROTOCOL_TLSv1_2)
            self.assertEqual(sslctx.verify_mode, ssl.CERT_REQUIRED)
            self.assertIs(sslctx.check_hostname, True)

        # Anonymous client sslconfig:
        sslconfig = pki.anonymous_client_sslconfig
        self.assertEqual(set(sslconfig), {'ca_file', 'check_hostname'})
        self.assertIs(sslconfig['check_hostname'], False)
        for func in client_sslctx_funcs:
            sslctx = func(sslconfig)
            self.assertIsInstance(sslctx, ssl.SSLContext)
            self.assertEqual(sslctx.protocol, ssl.PROTOCOL_TLSv1_2)
            self.assertEqual(sslctx.verify_mode, ssl.CERT_REQUIRED)
            self.assertIs(sslctx.check_hostname, False)

        # check_hostname should default to True:
        del sslconfig['check_hostname']
        for func in client_sslctx_funcs:
            sslctx = func(sslconfig)
            self.assertIsInstance(sslctx, ssl.SSLContext)
            self.assertEqual(sslctx.protocol, ssl.PROTOCOL_TLSv1_2)
            self.assertEqual(sslctx.verify_mode, ssl.CERT_REQUIRED)
            self.assertIs(sslctx.check_hostname, True)

    def test__validate_request(self):
        # Pre-build all valid types of length-encoded bodies:
        data = os.urandom(17)
        length_bodies = (
            data,
            bytearray(data),
            base.Body(io.BytesIO(data), 17),
            base.BodyIter([data], 17),
        )

        # Pre-build all valid types of chunk-encoded bodies:
        chunked_bodies = (
            base.ChunkedBody(io.BytesIO()),
            base.ChunkedBodyIter([])
        )

        # Bad method:
        with self.assertRaises(ValueError) as cm:
            client._validate_request(base.bodies, 'get', '/foo', {}, None)
        self.assertEqual(str(cm.exception), "invalid method: 'get'")

        # Non-casefolded header name:
        headers = {'Content-Type': 'text/plain', 'X-Stuff': 'hello'}
        with self.assertRaises(ValueError) as cm:
            client._validate_request(base.bodies, 'GET', '/foo', headers, None)
        self.assertEqual(str(cm.exception),
            "non-casefolded header name: 'Content-Type'"
        )

        # Body is None but content-length header included:
        headers = {'content-length': 17}
        with self.assertRaises(ValueError) as cm:
            client._validate_request(base.bodies, 'POST', '/foo', headers, None)
        self.assertEqual(str(cm.exception),
            "headers['content-length'] included when body is None"
        )

        # Body is None but transfer-encoding header included:
        headers = {'transfer-encoding': 'chunked'}
        with self.assertRaises(ValueError) as cm:
            client._validate_request(base.bodies, 'POST', '/foo', headers, None)
        self.assertEqual(str(cm.exception),
            "headers['transfer-encoding'] included when body is None"
        )

        # Bad body type:
        headers = {'content-type': 'text/plain'}
        with self.assertRaises(TypeError) as cm:
            client._validate_request(base.bodies, 'GET', '/foo', headers, 'hello')
        self.assertEqual(str(cm.exception),
            "bad request body type: <class 'str'>"
        )

        # 'transfer-encoding' header with a length-encoded body:
        for body in length_bodies:
            headers = {'transfer-encoding': 'chunked'}
            with self.assertRaises(ValueError) as cm:
                client._validate_request(base.bodies, 'POST', '/foo', headers, body)
            self.assertEqual(str(cm.exception),
                "headers['transfer-encoding'] with length-encoded body"
            )
            self.assertEqual(headers,
                {'content-length': 17, 'transfer-encoding': 'chunked'}
            )

        # headers['content-length'] != len(body):
        for body in length_bodies:
            headers = {'content-length': 16}
            with self.assertRaises(ValueError) as cm:
                client._validate_request(base.bodies, 'POST', '/foo', headers, body)
            self.assertEqual(str(cm.exception),
                "headers['content-length'] != len(body): 16 != 17"
            )
            self.assertEqual(headers, {'content-length': 16})

        # 'content-length' header with a chunk-encoded body:
        for body in chunked_bodies:
            headers = {'content-length': 5}
            with self.assertRaises(ValueError) as cm:
                client._validate_request(base.bodies, 'POST', '/foo', headers, body)
            self.assertEqual(str(cm.exception),
                "headers['content-length'] with chunk-encoded body"
            )
            self.assertEqual(headers,
                {'content-length': 5, 'transfer-encoding': 'chunked'}
            )

        # headers['transfer-encoding'] != 'chunked':
        for body in chunked_bodies:
            headers = {'transfer-encoding': 'clumped'}
            with self.assertRaises(ValueError) as cm:
                client._validate_request(base.bodies, 'POST', '/foo', headers, body)
            self.assertEqual(str(cm.exception),
                "headers['transfer-encoding'] is invalid: 'clumped'"
            )
            self.assertEqual(headers, {'transfer-encoding': 'clumped'})

        # Cannot include body when method is GET, HEAD, or DELETE:
        for body in (length_bodies + chunked_bodies):
            for method in ('GET', 'HEAD', 'DELETE'):
                with self.assertRaises(ValueError) as cm:
                    client._validate_request(base.bodies, method, '/foo', {}, body)
                self.assertEqual(str(cm.exception),
                    'cannot include body in a {} request'.format(method)
                )

        # No in-place header modification should happen when body is None:
        for method in ('GET', 'HEAD', 'DELETE', 'PUT', 'POST'):
            headers = {}
            self.assertIsNone(
                client._validate_request(base.bodies, method, '/foo', headers, None)
            )
            self.assertEqual(headers, {})

        # Test with all the length-encoded body types:
        for method in ('PUT', 'POST'):
            for body in length_bodies:
                headers = {}
                self.assertIsNone(
                    client._validate_request(base.bodies, method, '/foo', headers, body)
                )
                self.assertEqual(headers, {'content-length': 17})

        # Test with all the chunk-encoded body types:
        for method in ('PUT', 'POST'):
            for body in chunked_bodies:
                headers = {}
                self.assertIsNone(
                    client._validate_request(base.bodies, method, '/foo', headers, body)
                )
                self.assertEqual(headers, {'transfer-encoding': 'chunked'})

    def test__write_request(self):
        # Empty headers, no body:
        wfile = io.BytesIO()
        self.assertEqual(
            client._write_request(wfile, 'GET', '/', {}, None),
            18
        )
        self.assertEqual(wfile.tell(), 18)
        self.assertEqual(wfile.getvalue(), b'GET / HTTP/1.1\r\n\r\n')

        # One header:
        headers = {'foo': 17}  # Make sure to test with int header value
        wfile = io.BytesIO()
        self.assertEqual(
            client._write_request(wfile, 'GET', '/', headers, None),
            27
        )
        self.assertEqual(wfile.tell(), 27)
        self.assertEqual(wfile.getvalue(),
            b'GET / HTTP/1.1\r\nfoo: 17\r\n\r\n'
        )

        # Two headers:
        headers = {'foo': 17, 'bar': 'baz'}
        wfile = io.BytesIO()
        self.assertEqual(
            client._write_request(wfile, 'GET', '/', headers, None),
            37
        )
        self.assertEqual(wfile.tell(), 37)
        self.assertEqual(wfile.getvalue(),
            b'GET / HTTP/1.1\r\nbar: baz\r\nfoo: 17\r\n\r\n'
        )

        # body is bytes:
        wfile = io.BytesIO()
        self.assertEqual(
            client._write_request(wfile, 'GET', '/', headers, b'hello'),
            42
        )
        self.assertEqual(wfile.tell(), 42)
        self.assertEqual(wfile.getvalue(),
            b'GET / HTTP/1.1\r\nbar: baz\r\nfoo: 17\r\n\r\nhello'
        )

        # body is bytearray:
        body = bytearray(b'hello')
        wfile = io.BytesIO()
        self.assertEqual(
            client._write_request(wfile, 'GET', '/', headers, body),
            42
        )
        self.assertEqual(wfile.tell(), 42)
        self.assertEqual(wfile.getvalue(),
            b'GET / HTTP/1.1\r\nbar: baz\r\nfoo: 17\r\n\r\nhello'
        )

        # body is base.Body:
        rfile = io.BytesIO(b'hello')
        body = base.Body(rfile, 5)
        wfile = io.BytesIO()
        self.assertEqual(
            client._write_request(wfile, 'GET', '/', headers, body),
            42
        )
        self.assertEqual(rfile.tell(), 5)
        self.assertEqual(wfile.tell(), 42)
        self.assertEqual(wfile.getvalue(),
            b'GET / HTTP/1.1\r\nbar: baz\r\nfoo: 17\r\n\r\nhello'
        )

        # body is base.ChunkedBody:
        rfile = io.BytesIO(b'5\r\nhello\r\n0\r\n\r\n')
        body = base.ChunkedBody(rfile)
        wfile = io.BytesIO()
        self.assertEqual(
            client._write_request(wfile, 'GET', '/', headers, body),
            52
        )
        self.assertEqual(rfile.tell(), 15)
        self.assertEqual(wfile.tell(), 52)
        self.assertEqual(wfile.getvalue(),
            b'GET / HTTP/1.1\r\nbar: baz\r\nfoo: 17\r\n\r\n5\r\nhello\r\n0\r\n\r\n'
        )

    def test__read_response(self):
        # No headers, no body:
        lines = ''.join([
            'HTTP/1.1 200 OK\r\n',
            '\r\n',
        ]).encode('latin_1')
        rfile = io.BytesIO(lines)
        r = client._read_response(rfile, base.bodies, 'GET')
        self.assertIsInstance(r, client.Response)
        self.assertEqual(r, (200, 'OK', {}, None))

        # Content-Length, body should be base.Body:
        lines = ''.join([
            'HTTP/1.1 200 OK\r\n',
            'Content-Length: 17\r\n',
            '\r\n',
        ]).encode('latin_1')
        data = os.urandom(17)
        rfile = io.BytesIO(lines + data)
        r = client._read_response(rfile, base.bodies, 'GET')
        self.assertIsInstance(r, client.Response)
        self.assertEqual(r.status, 200)
        self.assertEqual(r.reason, 'OK')
        self.assertEqual(r.headers, {'content-length': 17})
        self.assertIsInstance(r.body, base.Body)
        self.assertIs(r.body.rfile, rfile)
        self.assertIs(r.body.closed, False)
        self.assertEqual(r.body._remaining, 17)
        self.assertEqual(rfile.tell(), len(lines))
        self.assertEqual(r.body.read(), data)
        self.assertEqual(rfile.tell(), len(lines) + len(data))
        self.assertIs(r.body.closed, True)
        self.assertEqual(r.body._remaining, 0)

        # Like above, except this time for a HEAD request:
        rfile = io.BytesIO(lines + data)
        r = client._read_response(rfile, base.bodies, 'HEAD')
        self.assertIsInstance(r, client.Response)
        self.assertEqual(r, (200, 'OK', {'content-length': 17}, None))

        # Transfer-Encoding, body should be base.ChunkedBody:
        lines = ''.join([
            'HTTP/1.1 200 OK\r\n',
            'Transfer-Encoding: chunked\r\n',
            '\r\n',
        ]).encode('latin_1')
        chunk1 = os.urandom(21)
        chunk2 = os.urandom(17)
        chunk3 = os.urandom(19)
        rfile = io.BytesIO()
        total = rfile.write(lines)
        for chunk in [chunk1, chunk2, chunk3, b'']:
            total += base.write_chunk(rfile, (None, chunk))
        self.assertEqual(rfile.tell(), total)
        rfile.seek(0)
        r = client._read_response(rfile, base.bodies, 'GET')
        self.assertIsInstance(r, client.Response)
        self.assertEqual(r.status, 200)
        self.assertEqual(r.reason, 'OK')
        self.assertEqual(r.headers, {'transfer-encoding': 'chunked'})
        self.assertIsInstance(r.body, base.ChunkedBody)
        self.assertIs(r.body.rfile, rfile)
        self.assertEqual(rfile.tell(), len(lines))
        self.assertIs(r.body.closed, False)
        self.assertEqual(list(r.body),
            [
                (None, chunk1),
                (None, chunk2),
                (None, chunk3),
                (None, b''),
            ]
        )
        self.assertIs(r.body.closed, True)
        self.assertEqual(rfile.tell(), total)


class TestConnection(TestCase):
    def test_init(self):
        sock = DummySocket()
        base_headers = {'host': 'www.example.com:80'}
        inst = client.Connection(sock, base_headers, base.bodies)
        self.assertIsInstance(inst, client.Connection)
        self.assertIs(inst.sock, sock)
        self.assertIs(inst.base_headers, base_headers)
        self.assertEqual(inst.base_headers, {'host': 'www.example.com:80'})
        self.assertIs(inst.bodies, base.bodies)
        self.assertIs(inst._rfile, sock._rfile)
        self.assertIs(inst._wfile, sock._wfile)
        self.assertIsNone(inst._response_body)
        self.assertIs(inst.closed, False)
        self.assertEqual(sock._calls, [
            ('makefile', 'rb', {'buffering': base.STREAM_BUFFER_SIZE}),
            ('makefile', 'wb', {'buffering': base.STREAM_BUFFER_SIZE}),
        ])

    def test_del(self):
        class ConnectionSubclass(client.Connection):
            def __init__(self):
                self._calls = 0

            def close(self):
                self._calls += 1

        inst = ConnectionSubclass()
        self.assertEqual(inst._calls, 0)
        self.assertIsNone(inst.__del__())
        self.assertEqual(inst._calls, 1)
        self.assertIsNone(inst.__del__())
        self.assertEqual(inst._calls, 2)

    def test_close(self):
        sock = DummySocket()
        inst = client.Connection(sock, None, base.bodies)
        sock._calls.clear()

        # When Connection.closed is False:
        self.assertIsNone(inst.close())
        self.assertEqual(sock._calls, [('shutdown', socket.SHUT_RDWR)])
        self.assertIsNone(inst.sock)
        self.assertIsNone(inst._response_body)
        self.assertIs(inst.closed, True)

        # Now when Connection.closed is True:
        self.assertIsNone(inst.close())
        self.assertEqual(sock._calls, [('shutdown', socket.SHUT_RDWR)])
        self.assertIsNone(inst.sock)
        self.assertIsNone(inst._response_body)
        self.assertIs(inst.closed, True)

    def test_request(self):
        # Test when the connection has already been closed:
        sock = DummySocket()
        conn = client.Connection(sock, None, base.bodies)
        sock._calls.clear()
        conn.sock = None
        with self.assertRaises(client.ClosedConnectionError) as cm:
            conn.request('GET', '/', {}, None)
        self.assertIs(cm.exception.conn, conn)
        self.assertEqual(str(cm.exception),
            'cannot use request() when connection is closed: {!r}'.format(conn)
        )
        self.assertEqual(sock._calls, [])
        self.assertIsNone(conn.sock)
        self.assertIsNone(conn._response_body)
        self.assertIs(conn.closed, True)

        # Test when the previous response body wasn't consumed:
        class DummyBody:
            closed = False

        sock = DummySocket()
        conn = client.Connection(sock, None, base.bodies)
        sock._calls.clear()
        conn._response_body = DummyBody
        with self.assertRaises(client.UnconsumedResponseError) as cm:
            conn.request('GET', '/', {}, None)
        self.assertIs(cm.exception.body, DummyBody)
        self.assertEqual(str(cm.exception),
            'previous response body not consumed: {!r}'.format(DummyBody)
        )
        # Make sure Connection.close() was called:
        self.assertEqual(sock._calls, [('shutdown', socket.SHUT_RDWR)])
        self.assertIsNone(conn.sock)
        self.assertIsNone(conn._response_body)
        self.assertIs(conn.closed, True)

    def get_subclass(self, marker):
        class Subclass(client.Connection):
            def __init__(self, marker):
                self._marker = marker
                self._args = None
                self._kw = None

            def request(self, *args, **kw):
                assert self._args is None
                assert self._kw is None
                self._args = args
                self._kw = kw
                return self._marker

        return Subclass(marker)

    def test_put(self):
        marker = random_id()
        inst = self.get_subclass(marker)
        uri = random_id()
        key = random_id().lower()
        value = random_id()
        headers = {key: value}
        body = os.urandom(16)
        self.assertIs(inst.put(uri, headers, body), marker)
        self.assertEqual(inst._args, ('PUT', uri, {key: value}, body))
        self.assertEqual(inst._kw, {})

    def test_post(self):
        marker = random_id()
        inst = self.get_subclass(marker)
        uri = random_id()
        key = random_id().lower()
        value = random_id()
        headers = {key: value}
        body = os.urandom(16)
        self.assertIs(inst.post(uri, headers, body), marker)
        self.assertEqual(inst._args, ('POST', uri, {key: value}, body))
        self.assertEqual(inst._kw, {})

    def test_get(self):
        marker = random_id()
        inst = self.get_subclass(marker)
        uri = random_id()
        key = random_id().lower()
        value = random_id()
        headers = {key: value}
        self.assertIs(inst.get(uri, headers), marker)
        self.assertEqual(inst._args, ('GET', uri, {key: value}, None))
        self.assertEqual(inst._kw, {})

    def test_head(self):
        marker = random_id()
        inst = self.get_subclass(marker)
        uri = random_id()
        key = random_id().lower()
        value = random_id()
        headers = {key: value}
        self.assertIs(inst.head(uri, headers), marker)
        self.assertEqual(inst._args, ('HEAD', uri, {key: value}, None))
        self.assertEqual(inst._kw, {})

    def test_delete(self):
        marker = random_id()
        inst = self.get_subclass(marker)
        uri = random_id()
        key = random_id().lower()
        value = random_id()
        headers = {key: value}
        self.assertIs(inst.delete(uri, headers), marker)
        self.assertEqual(inst._args, ('DELETE', uri, {key: value}, None))
        self.assertEqual(inst._kw, {})   


class TestClient(TestCase):
    def test_init(self):
        # Bad address type:
        with self.assertRaises(TypeError) as cm:
            client.Client(1234)
        self.assertEqual(str(cm.exception),
            _TYPE_ERROR.format('address', (tuple, str, bytes), int, 1234)
        )

        # Wrong number of items in address tuple:
        for address in BAD_TUPLE_ADDRESSES:
            self.assertIn(len(address), {1, 3, 5})
            with self.assertRaises(ValueError) as cm:
                client.Client(address)
            self.assertEqual(str(cm.exception),
                'address: must have 2 or 4 items; got {!r}'.format(address)
            )

        # Non-absolute/non-normalized AF_UNIX filename:
        with self.assertRaises(ValueError) as cm:
            client.Client('foo')
        self.assertEqual(str(cm.exception),
            "address: bad socket filename: 'foo'"
        )

        # Good address type and value permutations:
        for address in GOOD_ADDRESSES:
            inst = client.Client(address)
            self.assertIs(inst.address, address)
            self.assertEqual(inst.options, {})
            if isinstance(address, tuple):
                self.assertEqual(inst.host, client._build_host(80, *address))
            else:
                self.assertIsNone(inst.host)
            self.assertEqual(inst.timeout, 60)
            self.assertIs(inst.bodies, base.bodies)
            self.assertIsNone(inst.on_connect)

            # Test overriding the `host` option:
            my_host = '.'.join([random_id() for i in range(3)])
            inst = client.Client(address, host=my_host)
            self.assertIs(inst.address, address)
            self.assertEqual(inst.options, {'host': my_host})
            self.assertIs(inst.host, my_host)
            self.assertEqual(inst.timeout, 60)
            self.assertIs(inst.bodies, base.bodies)
            self.assertIsNone(inst.on_connect)

            # Test overriding `host` option with `None`:
            inst = client.Client(address, host=None)
            self.assertIs(inst.address, address)
            self.assertEqual(inst.options, {'host': None})
            self.assertIsNone(inst.host)
            self.assertEqual(inst.timeout, 60)
            self.assertIs(inst.bodies, base.bodies)
            self.assertIsNone(inst.on_connect)

            # Test overriding the `timeout` option:
            inst = client.Client(address, timeout=17)
            self.assertIs(inst.address, address)
            self.assertEqual(inst.options, {'timeout': 17})
            if isinstance(address, tuple):
                self.assertEqual(inst.host, client._build_host(80, *address))
            else:
                self.assertIsNone(inst.host)
            self.assertEqual(inst.timeout, 17)
            self.assertIs(inst.bodies, base.bodies)
            self.assertIsNone(inst.on_connect)

            # Test overriding the `bodies` option:
            my_bodies = base.BodiesAPI('foo', 'bar', 'stuff', 'junk')
            inst = client.Client(address, bodies=my_bodies)
            self.assertIs(inst.address, address)
            self.assertEqual(inst.options, {'bodies': my_bodies})
            if isinstance(address, tuple):
                self.assertEqual(inst.host, client._build_host(80, *address))
            else:
                self.assertIsNone(inst.host)
            self.assertEqual(inst.timeout, 60)
            self.assertIs(inst.bodies, my_bodies)
            self.assertIsNone(inst.on_connect)

            # Test overriding the `on_connect` option:
            inst = client.Client(address, on_connect=None)
            if isinstance(address, tuple):
                self.assertEqual(inst.host, client._build_host(80, *address))
            else:
                self.assertIsNone(inst.host)
            self.assertEqual(inst.timeout, 60)
            self.assertIs(inst.bodies, base.bodies)
            self.assertIsNone(inst.on_connect)

            with self.assertRaises(TypeError) as cm:
                client.Client(address, on_connect='hello')
            self.assertEqual(str(cm.exception),
                "on_connect: not callable: 'hello'"
            )

            def my_on_connect(conn):
                return True
            inst = client.Client(address, on_connect=my_on_connect)
            if isinstance(address, tuple):
                self.assertEqual(inst.host, client._build_host(80, *address))
            else:
                self.assertIsNone(inst.host)
            self.assertEqual(inst.timeout, 60)
            self.assertIs(inst.bodies, base.bodies)
            self.assertIs(inst.on_connect, my_on_connect)

            # Test overriding all the options together:
            options = {
                'host': my_host,
                'timeout': 16.9,
                'bodies': my_bodies,
                'on_connect': my_on_connect,
            }
            inst = client.Client(address, **options)
            self.assertIs(inst.address, address)
            self.assertEqual(inst.options, options)
            self.assertIs(inst.host, my_host)
            self.assertEqual(inst.timeout, 16.9)
            self.assertIs(inst.bodies, my_bodies)
            self.assertIs(inst.on_connect, my_on_connect)

    def test_repr(self):
        class Custom(client.Client):
            pass

        for address in GOOD_ADDRESSES:
            inst = client.Client(address)
            self.assertEqual(repr(inst), 'Client({!r})'.format(address))
            inst = Custom(address)
            self.assertEqual(repr(inst), 'Custom({!r})'.format(address))

    def test_connect(self):
        class ClientSubclass(client.Client):
            def __init__(self, sock, host, on_connect=None):
                self.__sock = sock
                self._base_headers = {'host': host}
                self.bodies = base.bodies
                self.on_connect = on_connect

            def create_socket(self):
                return self.__sock

        sock = DummySocket()
        host = random_id().lower()
        inst = ClientSubclass(sock, host)
        self.assertIsNone(inst.on_connect)
        conn = inst.connect()
        self.assertIsInstance(conn, client.Connection)
        self.assertIs(conn.sock, sock)
        self.assertIs(conn.base_headers, inst._base_headers)
        self.assertIs(conn.bodies, base.bodies)
        self.assertIs(conn._rfile, sock._rfile)
        self.assertIs(conn._wfile, sock._wfile)
        self.assertEqual(sock._calls, [
            ('makefile', 'rb', {'buffering': base.STREAM_BUFFER_SIZE}),
            ('makefile', 'wb', {'buffering': base.STREAM_BUFFER_SIZE}),
        ])

        # Should return a new Connection instance each time:
        conn2 = inst.connect()
        self.assertIsNot(conn2, conn)
        self.assertIsInstance(conn2, client.Connection)
        self.assertIs(conn2.sock, sock)
        self.assertIs(conn.base_headers, inst._base_headers)
        self.assertIs(conn.bodies, base.bodies)
        self.assertIs(conn2._rfile, sock._rfile)
        self.assertIs(conn2._wfile, sock._wfile)
        self.assertEqual(sock._calls, [
            ('makefile', 'rb', {'buffering': base.STREAM_BUFFER_SIZE}),
            ('makefile', 'wb', {'buffering': base.STREAM_BUFFER_SIZE}),
            ('makefile', 'rb', {'buffering': base.STREAM_BUFFER_SIZE}),
            ('makefile', 'wb', {'buffering': base.STREAM_BUFFER_SIZE}),
        ])

        # on_connect() returns True:
        def on_connect_true(conn):
            return True
        sock = DummySocket()
        host = random_id().lower()
        inst = ClientSubclass(sock, host, on_connect_true)
        self.assertIs(inst.on_connect, on_connect_true)
        conn = inst.connect()
        self.assertIsInstance(conn, client.Connection)
        self.assertIs(conn.sock, sock)
        self.assertIs(conn.base_headers, inst._base_headers)
        self.assertIs(conn.bodies, base.bodies)
        self.assertIs(conn._rfile, sock._rfile)
        self.assertIs(conn._wfile, sock._wfile)
        self.assertEqual(sock._calls, [
            ('makefile', 'rb', {'buffering': base.STREAM_BUFFER_SIZE}),
            ('makefile', 'wb', {'buffering': base.STREAM_BUFFER_SIZE}),
        ])

        # on_connect() does not return True:
        def on_connect_false(conn):
            return 1
        sock = DummySocket()
        host = random_id().lower()
        inst = ClientSubclass(sock, host, on_connect_false)
        self.assertIs(inst.on_connect, on_connect_false)
        with self.assertRaises(ValueError) as cm:
            conn = inst.connect()
        self.assertEqual(str(cm.exception), 'on_connect() did not return True')


class TestSSLClient(TestCase):
    def test_init(self):
        # sslctx is not an ssl.SSLContext:
        with self.assertRaises(TypeError) as cm:
            client.SSLClient('foo', None)
        self.assertEqual(str(cm.exception), 'sslctx must be an ssl.SSLContext')

        # Bad SSL protocol version:
        sslctx = ssl.SSLContext(ssl.PROTOCOL_SSLv3)
        with self.assertRaises(ValueError) as cm:
            client.SSLClient(sslctx, None)
        self.assertEqual(str(cm.exception),
            'sslctx.protocol must be ssl.PROTOCOL_TLSv1_2'
        )

        # Note: Python 3.3.4 (and presumably 3.4.0) now disables SSLv2 by
        # default (which is good); Degu enforces this (also good), but because
        # we cannot unset the ssl.OP_NO_SSLv2 bit, we can't unit test to check
        # that Degu enforces this, so for now, we set the bit here so it works
        # with Python 3.3.3 still; see: http://bugs.python.org/issue20207
        sslctx = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        sslctx.options |= ssl.OP_NO_SSLv2

        # not (options & ssl.OP_NO_COMPRESSION)
        sslctx.options |= ssl.OP_NO_SSLv2
        with self.assertRaises(ValueError) as cm:
            client.SSLClient(sslctx, None)
        self.assertEqual(str(cm.exception),
            'sslctx.options must include ssl.OP_NO_COMPRESSION'
        )

        # verify_mode is not ssl.CERT_REQUIRED:
        sslctx.options |= ssl.OP_NO_COMPRESSION
        with self.assertRaises(ValueError) as cm:
            client.SSLClient(sslctx, None)
        self.assertEqual(str(cm.exception),
            'sslctx.verify_mode must be ssl.CERT_REQUIRED'
        )

        #############################
        # Good sslctx from here on...
        sslctx.verify_mode = ssl.CERT_REQUIRED

        # Bad address type:
        with self.assertRaises(TypeError) as cm:
            client.SSLClient(sslctx, 1234)
        self.assertEqual(str(cm.exception),
            _TYPE_ERROR.format('address', (tuple, str, bytes), int, 1234)
        )

        # Wrong number of items in address tuple:
        for address in BAD_TUPLE_ADDRESSES:
            self.assertIn(len(address), {1, 3, 5})
            with self.assertRaises(ValueError) as cm:
                client.SSLClient(sslctx, address)
            self.assertEqual(str(cm.exception),
                'address: must have 2 or 4 items; got {!r}'.format(address)
            )

        # Non-absolute/non-normalized AF_UNIX filename:
        with self.assertRaises(ValueError) as cm:
            client.SSLClient(sslctx, 'foo')
        self.assertEqual(str(cm.exception),
            "address: bad socket filename: 'foo'"
        )

        # Good address type and value permutations:
        for address in GOOD_ADDRESSES:
            inst = client.SSLClient(sslctx, address)
            self.assertIs(inst.address, address)
            self.assertEqual(inst.options, {})
            if isinstance(address, tuple):
                self.assertEqual(inst.host, client._build_host(443, *address))
                self.assertIs(inst.ssl_host, address[0])
            else:
                self.assertIsNone(inst.host)
                self.assertIsNone(inst.ssl_host)
            self.assertEqual(inst.timeout, 60)
            self.assertIs(inst.bodies, base.bodies)
            self.assertIsNone(inst.on_connect)

            # Test overriding the `host` option:
            my_host = '.'.join([random_id() for i in range(3)])
            inst = client.SSLClient(sslctx, address, host=my_host)
            self.assertIs(inst.address, address)
            self.assertEqual(inst.options, {'host': my_host})
            self.assertIs(inst.host, my_host)
            self.assertEqual(inst.timeout, 60)
            self.assertIs(inst.bodies, base.bodies)
            self.assertIsNone(inst.on_connect)

            # Test overriding the `host` option with `None`:
            inst = client.SSLClient(sslctx, address, host=None)
            self.assertIs(inst.address, address)
            self.assertEqual(inst.options, {'host': None})
            self.assertIsNone(inst.host)
            self.assertEqual(inst.timeout, 60)
            self.assertIs(inst.bodies, base.bodies)
            self.assertIsNone(inst.on_connect)

            # Test overriding the `ssl_host` options:
            my_ssl_host = '.'.join([random_id(10) for i in range(4)])
            inst = client.SSLClient(sslctx, address, ssl_host=my_ssl_host)
            self.assertIs(inst.address, address)
            self.assertEqual(inst.options, {'ssl_host': my_ssl_host})
            self.assertIs(inst.ssl_host, my_ssl_host)
            self.assertEqual(inst.timeout, 60)
            self.assertIs(inst.bodies, base.bodies)
            self.assertIsNone(inst.on_connect)

            # Test overriding the `ssl_host` option with `None`:
            inst = client.SSLClient(sslctx, address, ssl_host=None)
            self.assertIs(inst.address, address)
            self.assertEqual(inst.options, {'ssl_host': None})
            self.assertIsNone(inst.ssl_host)
            self.assertEqual(inst.timeout, 60)
            self.assertIs(inst.bodies, base.bodies)
            self.assertIsNone(inst.on_connect)

            # Test overriding the `timeout` option:
            inst = client.SSLClient(sslctx, address, timeout=17)
            self.assertIs(inst.address, address)
            self.assertEqual(inst.options, {'timeout': 17})
            if isinstance(address, tuple):
                self.assertEqual(inst.host, client._build_host(443, *address))
                self.assertIs(inst.ssl_host, address[0])
            else:
                self.assertIsNone(inst.host)
                self.assertIsNone(inst.ssl_host)
            self.assertEqual(inst.timeout, 17)
            self.assertIs(inst.bodies, base.bodies)
            self.assertIsNone(inst.on_connect)

            # Test overriding the `bodies` option:
            my_bodies = base.BodiesAPI('foo', 'bar', 'stuff', 'junk')
            inst = client.SSLClient(sslctx, address, bodies=my_bodies)
            self.assertIs(inst.address, address)
            self.assertEqual(inst.options, {'bodies': my_bodies})
            if isinstance(address, tuple):
                self.assertEqual(inst.host, client._build_host(443, *address))
                self.assertIs(inst.ssl_host, address[0])
            else:
                self.assertIsNone(inst.host)
                self.assertIsNone(inst.ssl_host)
            self.assertEqual(inst.timeout, 60)
            self.assertIs(inst.bodies, my_bodies)
            self.assertIsNone(inst.on_connect)

            # Test overriding the `on_connect` option:
            with self.assertRaises(TypeError) as cm:
                client.SSLClient(sslctx, address, on_connect='hello')
            self.assertEqual(str(cm.exception),
                "on_connect: not callable: 'hello'"
            )
            def my_on_connect(conn):
                return True
            inst = client.SSLClient(sslctx, address, on_connect=my_on_connect)
            if isinstance(address, tuple):
                self.assertEqual(inst.host, client._build_host(443, *address))
                self.assertIs(inst.ssl_host, address[0])
            else:
                self.assertIsNone(inst.host)
                self.assertIsNone(inst.ssl_host)
            self.assertEqual(inst.timeout, 60)
            self.assertIs(inst.bodies, base.bodies)
            self.assertIs(inst.on_connect, my_on_connect)

            # Test overriding the `on_connect` option with `None`:
            inst = client.SSLClient(sslctx, address, on_connect=None)
            if isinstance(address, tuple):
                self.assertEqual(inst.host, client._build_host(443, *address))
                self.assertIs(inst.ssl_host, address[0])
            else:
                self.assertIsNone(inst.host)
                self.assertIsNone(inst.ssl_host)
            self.assertEqual(inst.timeout, 60)
            self.assertIs(inst.bodies, base.bodies)
            self.assertIsNone(inst.on_connect)

            # Test overriding all the options together:
            options = {
                'host': my_host,
                'ssl_host': my_ssl_host,
                'timeout': 16.9,
                'bodies': my_bodies,
                'on_connect': my_on_connect,
            }
            inst = client.SSLClient(sslctx, address, **options)
            self.assertIs(inst.address, address)
            self.assertEqual(inst.options, options)
            self.assertIs(inst.host, my_host)
            self.assertIs(inst.ssl_host, my_ssl_host)
            self.assertEqual(inst.timeout, 16.9)
            self.assertIs(inst.bodies, my_bodies)
            self.assertIs(inst.on_connect, my_on_connect)

    def test_repr(self):
        class Custom(client.SSLClient):
            pass

        pki = TempPKI()
        sslctx = client.build_client_sslctx(pki.client_sslconfig)

        for address in GOOD_ADDRESSES:
            inst = client.SSLClient(sslctx, address)
            self.assertEqual(repr(inst),
                'SSLClient({!r}, {!r})'.format(sslctx, address)
            )
            inst = Custom(sslctx, address)
            self.assertEqual(repr(inst),
                'Custom({!r}, {!r})'.format(sslctx, address)
            )
